// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   11. Nov 2000
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: include file which handles ArtePointers
//
// changes:
// 11 Nov 2000 (TG) creation from C.Borgmeiers ArtePointer.hh code
// 11 Nov 2000 (TG) operator= added to class ArtePointer
// 16 Nov 2000 (TG) ClassDef only if USE_ROOT and USE_GATE
// 17 Nov 2000 (TG) added constructor to ArtePointer class
// 02 Apr 2001 (TG) added explicit to ArtePointer CTOR
//
// *****************************************************************************
#ifndef ARTEPTR_HH
#define ARTEPTR_HH
#include <iterator>
#include "arte/PointerBase.hh"

/** single link class
    @author C. Borgmeier
*/
//==============================================================================
// ArteLink
//==============================================================================
class ArteLink {
public:
  ///
  inline int& getRef() { return i; }
  ///
  inline int& getInt() { return i; }
#if !defined(ARTE_CXX_COMPAT_MODE) || ARTE_CXX_COMPAT_MODE==1
  /// only available in ARTE compatibility mode
  inline int operator ()() { return i; }
#endif

  ///
  inline void*& getVec() { return reinterpret_cast<void*&>( i ); }
protected:
  int i;
}; // end of ArteLink


/** Link vector
    @author C. Borgmeier
*/
//==============================================================================
// ArteVector
//==============================================================================
class ArteVector : public ArteLink {
public:
  ///
  inline const int & operator ()() { return getInt(); };
}; // end of ArteVector


/** Arte Relation Vector
    @author C. Borgmeier
*/
//==============================================================================
// ArteRelVector
//==============================================================================
template<class T>
class ArteRelVector : public ArteVector {
public:
  ///
  ArteRelVector();
  ///
  ~ArteRelVector();
  ///
  void clear();
}; // end of ArteRelVector

template <class T> class ArteTable;
template <class T> class ArteTableIterator;

/** Arte pointer class
    @author C. Borgmeier, T. Glebe
*/
//==============================================================================
// ArtePointer
//==============================================================================
template <class T>
class ArtePointer : public ArtePointerBase {
#ifndef __CINT__
  // CINT cannot resolve circular friendship between ArteTable && ArtePointer
  friend class ArteTable<T>;
  friend class ArteTableIterator<T>;
#endif

public:
  typedef ArtePointer<T>    self;

  typedef T                 value_type;
  typedef value_type*       pointer;
  typedef value_type&       reference;
#ifndef __CINT__
  typedef const value_type& const_reference;
#endif

  /** @name --- Constructors --- */
  ///
  ArtePointer(): ArtePointerBase( 0 ) {}
  ///
  ArtePointer( const UShort_t index ): ArtePointerBase( index ) {}
  /// this is a conversion between ArtePointers!!
  explicit ArtePointer( const ArtePointerBase& rhs): ArtePointerBase( rhs ) {}
  ///
  ArtePointer( const ArtePointer<T>& index ): ArtePointerBase( index ) {}

  ///
  const self& operator=(const self& rhs) {
    ArtePointerBase::operator=(rhs);
    return *this;
  }

  /** @name --- Operators --- */
  ///
  inline T& operator*() const { return *static_cast<T*>( dereference() ); }
  ///
  inline T* operator->() const { return static_cast<T*>( dereference() ); }

  ///
  inline bool operator> ( const ArtePointer<T>& rhs) const
    { return ArtePointerBase::operator>( rhs ); }
  ///
  inline bool operator< ( const ArtePointer<T>& rhs) const
    { return ArtePointerBase::operator<( rhs ); }
  ///
  inline int  operator- ( const ArtePointer<T>& rhs) const
    { return ArtePointerBase::operator-(rhs); }
  ///
  inline bool operator!=( const ArtePointer<T>& rhs ) const
    { return ArtePointerBase::operator!=( rhs ); }
  ///
  inline bool operator==( const ArtePointer<T>& rhs ) const
    { return ArtePointerBase::operator==( rhs ); }

#if !defined(ARTE_CXX_COMPAT_MODE) || (ARTE_CXX_COMPAT_MODE==1) || defined(CINTOBJECT)
  /** @name --- ARTE compatibility --- */
  ///
  inline ArtePointer<T>& operator++() { increment(); return *this; }
  ///
  inline ArtePointer<T>  operator++(int) {
    ArtePointer<T> tmp = *this;
    increment(); return tmp;
    //    return (ArtePointer<T>(*this) += 1);
  }

#ifndef NO_ARTE
  /** @name --- ARTE only functions --- */
  /// Returns the number of entries in the corresponding Arte table.
  inline static unsigned int nrow();
  ///
  inline static self append(const_reference e);
  ///
  inline static void*     getLink();
  ///
  inline static int&      getIntLink();
  ///
  inline static ArteLink& getLinkLink();

  /// Return the width of the Arte table.
  inline static int ncol();
#endif /* NO_ARTE */

public:
#else // ARTE_CXX_COMPAT
//  #ifndef __MAKECINT__
//    ArtePointer( Pointer* ): ArtePointerBase( 0 ) {}

//    inline bool operator!=( Pointer* ) const
//      { return ArtePointerBase::operator!=( 0 ); }
//  #endif
protected:
#endif /* COMPAT */
  // public only in ARTE compat. mode
  //  unsigned int index() const { return ArtePointerBase::index(); }

private:
  inline void* dereference() const;
  static int* getBase();

}; // End of class ArtePointer

//  template<class T>
//  inline void ArtePointer<T>::Streamer(TBuffer &R__b)
//  {
//    if (R__b.IsReading()) {
//      R__b >> *this;
//    } else {
//      R__b << *this;
//    }
//  }


/** Arte Relation Pointer
    @author C. Borgmeier
*/
//==============================================================================
// ArteRelationPointer
//==============================================================================
template<class Relation, class Result>
class ArteRelationPointer: public ArtePointerBase {
  typedef ArteRelationPointer<Relation,Result> self;
  typedef ArtePointer<Result> value_type;
  typedef value_type* pointer;
  typedef value_type& reference;

  friend class ArteRelVector<Relation>;
public:
  /** @name --- Constructors --- */
  ///
  ArteRelationPointer(): ArtePointerBase( 0 ) {}
  ///
  ArteRelationPointer( const ArteRelationPointer<Relation,Result>& arg )
    : ArtePointerBase( arg ) {}
}; // end of ArteRelationPointer


template<class Result> class ArteRelationIterator;


/** ArteInvOneRelationPointer
    @author C. Borgmeier
*/
//==============================================================================
// ArteInvOneRelationPointer
//==============================================================================
template<class Relation, class Result>
class ArteInvOneRelationPointer: public ArteRelationPointer<Relation,Result> {
  typedef ArteInvOneRelationPointer<Relation,Result> self;
  typedef ArtePointer<Result> value_type;
  typedef value_type* pointer;
  typedef value_type& reference;
  typedef ArteRelationIterator<Result> iterator;
  typedef std::forward_iterator_tag iterator_category;
  typedef int difference_type; // CINT does not know ptrdiff_t
  //  typedef ptrdiff_t difference_type;

public:  
  ///
  iterator begin() const;
  ///
  iterator end() const;
  ///
  inline static ArteRelVector<Relation>& getLink();
#if !defined(ARTE_CXX_COMPAT_MODE) || (ARTE_CXX_COMPAT_MODE==1)
  /// only in ARTE compat. mode
  inline static ArteRelVector<Relation>& getLinkLink()
    { return getLink(); }
#endif
}; // end of ArteInvOneRelationPointer


/** ArteManRelationPointer
    @author C. Borgmeier
*/
//==============================================================================
// ArteManRelationPointer
//==============================================================================
template<class Relation, class Result>
class ArteManRelationPointer: public ArteRelationPointer<Relation,Result> {
  typedef ArteManRelationPointer<Relation,Result> self;
  typedef ArtePointer<Result> value_type;
  typedef value_type* pointer;
  typedef value_type& reference;
  typedef ArteRelationIterator<Result> iterator;
  typedef std::forward_iterator_tag iterator_category;
  typedef int difference_type; // CINT does not know ptrdiff_t
  //  typedef ptrdiff_t difference_type;

public:
  ///
  iterator begin() const;
  ///
  iterator end() const;
  ///
  inline static ArteRelVector<Relation>& getLink();
#if !defined(ARTE_CXX_COMPAT_MODE) || (ARTE_CXX_COMPAT_MODE==1)
  /// only in ARTE compat. mode
  inline static ArteRelVector<Relation>& getLinkLink()
    { return getLink(); }
#endif
}; // end of ArteManRelationPointer


/** ArteInvManRelationPointer
    @author C. Borgmeier
*/
//==============================================================================
// ArteInvManRelationPointer
//==============================================================================
template<class Relation, class Result>
class ArteInvManRelationPointer: public ArteRelationPointer<Relation,Result> {
  typedef ArteInvManRelationPointer<Relation,Result> self;
  typedef ArtePointer<Result> value_type;
  typedef value_type* pointer;
  typedef value_type& reference;
  typedef ArteRelationIterator<Result> iterator;
  typedef std::forward_iterator_tag iterator_category;
  typedef int difference_type; // CINT does not know ptrdiff_t
  //  typedef ptrdiff_t difference_type;

public:
  ///
  iterator begin() const;
  ///
  iterator end() const;
  ///
  inline static ArteRelVector<Relation>& getLink();
}; // end of ArteInvManRelationPointer


/** ArteRelationIterator
    @author C. Borgmeier, T. Glebe
*/
//==============================================================================
// ArteRelationIterator
//==============================================================================
template<class Result>
class ArteRelationIterator: public ArtePointerBase {
public:
  typedef ArteRelationIterator<Result> self;

  typedef ArtePointer<Result> value_type;
  typedef value_type* pointer;
  typedef value_type& reference;
  typedef self iterator;
  typedef std::forward_iterator_tag iterator_category;
  typedef int difference_type; // CINT does not know ptrdiff_t

  /** @name --- Constructors --- */
  ///
  ArteRelationIterator();
  ///
  template<class Relation>
  ArteRelationIterator( const ArteInvOneRelationPointer<Relation,Result>& x )
    : ArtePointerBase( x ),
      relation( &x.getLink() ),
      relType( 2 )
    {}
  ///
  template<class Relation>
  ArteRelationIterator( const ArteManRelationPointer<Relation,Result>& x )
    : ArtePointerBase( x ),
      relation( &x.getLink() ),
      relType( 3 )
    {}
  ///
  template<class Relation>
  ArteRelationIterator( const ArteInvManRelationPointer<Relation,Result>& x )
    : ArtePointerBase( x ),
      relation( &x.getLink() ),
      relType( 4 )
    {}

  /** @name --- Relation structures --- */
  ///
  struct ArteOneToManyRelation {
    //    ArteOneToManyRelation(): next( ArtePointer<Result>(0) ) {}
    ArteOneToManyRelation(): next( ArtePointer<Result>() ) {}
    ArtePointer<Result> el;
    ArtePointerBase next;
  };
  
  ///
  struct ArteManyToManyRelation {
//      ArteManyToManyRelation()
//        : next(     ArtePointer<Result>( 0 ) ),
//    	othernext(ArtePointer<Result>( 0 ) ) {}
    ArteManyToManyRelation()
      : next      ( ArtePointer<Result>() ),
  	othernext ( ArtePointer<Result>() ) {}

    ArtePointer<Result> el;
    ArtePointerBase next;
    int selfel;  // we do not know the type
    ArtePointerBase othernext;
  }; // end of ArteManyToManyRelation

  ///
  struct ArteBackwardsManyToManyRelation {
//      ArteBackwardsManyToManyRelation()
//        : othernext( ArtePointer<Result>(0) ),
//  	next(      ArtePointer<Result>(0) ) {}
    ArteBackwardsManyToManyRelation()
      : othernext ( ArtePointer<Result>() ),
	next      ( ArtePointer<Result>() ) {}
    int selfel;  // we do not know the type
    ArtePointerBase othernext;
    ArtePointer<Result> el;
    ArtePointerBase next;
  }; // end of ArteBackwardsManyToManyRelation
  

  /** @name --- Operators --- */
  ///
  template<class Relation>
  const ArteRelationIterator<Result>&
  operator=( const ArteInvOneRelationPointer<Relation,Result>& x )
    {
      *(ArtePointerBase*)this = x;
      relation = &x.getLink();
      reltype  = 2;
      return *this;
    }

  ///
  template<class Relation>
  const ArteRelationIterator<Result>&
  operator=( const ArteManRelationPointer<Relation,Result>& x )
    {
      *(ArtePointerBase*)this = x;
      relation = &x.getLink();
      reltype  = 3;
      return *this;
    }

  ///
  template<class Relation>
  const ArteRelationIterator<Result>&
  operator=( const ArteInvManRelationPointer<Relation,Result>& x )
    {
      *(ArtePointerBase*)this = x;
      relation = &x.getLink();
      reltype  = 4;
      return *this;
    }

  ///
  inline ArtePointer<Result> operator*() const { return lookup(); }
  // Dereferences the pointer.

#if !defined(ARTE_CXX_COMPAT_MODE) || (ARTE_CXX_COMPAT_MODE==1)
  /// only in ARTE compat. mode
  operator ArtePointer<Result>() const { return lookup(); }
#endif

  ///
  self& operator++() {
    followLink();
    return *this;
  }
  ///
  self operator++(int) {
    self tmp = *this;
    followLink();
    return tmp;
  }
  ///
  inline bool operator!=( const ArteRelationIterator<Result>& rhs )
    { return ArtePointerBase::operator!=( rhs ); }
  ///
  inline bool operator==( const ArteRelationIterator<Result>& rhs )
    { return ArtePointerBase::operator==( rhs ); }

#if !defined(ARTE_CXX_COMPAT_MODE) || (ARTE_CXX_COMPAT_MODE==1)
  /** @name --- ARTE compatibility mode --- */
  ///
  inline int valid() const { return ArtePointerBase::operator!=(0); }
  ///
  inline void next() { followLink(); }
#endif
  
private:
  void* getBase() const
    {
#ifndef NO_ARTE
      return gcbank_.iq + (*relation).getInt() + 2;
#else
      cerr << "wrong getBase called" << endl;
      return 0;
#endif
    }

  ArteManyToManyRelation getManyToMany() const {
    ArteManyToManyRelation* rp
      = static_cast<ArteManyToManyRelation*>( getBase() );
    return rp[index()-1];      
  }
  
  ArteBackwardsManyToManyRelation getBackwardsManyToMany() const {
    ArteBackwardsManyToManyRelation* rp
      = static_cast<ArteBackwardsManyToManyRelation*>( getBase() );
    return rp[index()-1];
  }
  
  ArteOneToManyRelation getOneToMany() const {
    ArteOneToManyRelation* rp
      = static_cast<ArteOneToManyRelation*>( getBase() );
    return rp[index()-1];
  }

protected:
  ///
  ArtePointer<Result> lookup() const {
    switch( relType ) {
    case 2: { // OneToMany
      ArteOneToManyRelation rel = getOneToMany();
      return rel.el;
    }
    case 3: { // ManyToMany
      ArteManyToManyRelation rel = getManyToMany();
      return rel.el;
    }
    case 4:
    default: { // BackwardsManyToMany
      ArteBackwardsManyToManyRelation rel = getBackwardsManyToMany();
      return rel.el;
    }
    } // switch
  } // lookup()

  ///
  void followLink() {
    if( index()!= 0 ) {
      switch( relType ) {
	
      case 2: { // OneToMany
	ArteOneToManyRelation rel = getOneToMany();
	//indexReference() = rel.next.index();
	*static_cast<ArtePointerBase*>(this) = rel.next;
	break;
      }
      case 3: { // ManyToMany
	ArteManyToManyRelation rel = getManyToMany();
	*static_cast<ArtePointerBase*>(this) = rel.next;
	break;
      }
      case 4:
      default: { // ManyToMany
	ArteBackwardsManyToManyRelation rel
	  = getBackwardsManyToMany();
	*static_cast<ArtePointerBase*>(this) = rel.next;
	break;
      }
      } // switch
    } // if index()
  } // followLink()    

private:
  ArteLink* relation;
  int relType;
}; // End of class ArteRelationIterator


template<class Relation, class Result>
inline ArteRelationIterator<Result>
ArteInvOneRelationPointer<Relation,Result>::begin() const
{
  return ArteRelationIterator<Result>( *this );
}


template<class Relation, class Result>
inline ArteRelationIterator<Result>
ArteManRelationPointer<Relation,Result>::begin() const
{
  return ArteRelationIterator<Result>( *this );
}


template<class Relation, class Result>
inline ArteRelationIterator<Result>
ArteInvManRelationPointer<Relation,Result>::begin() const
{
  return ArteRelationIterator<Result>( *this );
}


template<class Relation, class Result>
inline ArteRelationIterator<Result>
ArteInvManRelationPointer<Relation,Result>::end() const
{
  return ArteRelationIterator<Result>();
}


template<class Relation, class Result>
inline ArteRelationIterator<Result>
ArteManRelationPointer<Relation,Result>::end() const
{
  return ArteRelationIterator<Result>();
}


template<class Relation, class Result>
inline ArteRelationIterator<Result>
ArteInvOneRelationPointer<Relation,Result>::end() const
{
  return ArteRelationIterator<Result>();
}
#endif
