// **********************************************************************
//
// source:
//
// type:      source code
//
// created:   20. Jul 1997
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Class to contain the ROOT Event
//
// changes:
// 20 Jul 1997 creation by C. Borgmeier
// 30 Okt 2000 (TG) removed fillWireBuffer(), fillTrackBuffer(),
//             findPrimVertices(), findSecVertices()
// 20 Nov 2000 (TG) added namespace CLUE
// 18 Dec 2000 (TG) bug fix in DTOR: addeded deletion of muon-container
// 19 Dec 2000 (TG) bug fix in Clear(): muon->Clear() added
// 22 Dec 2000 (TG) bug fix in Clear(): containers of classes using
//                  relation containers must use Delete() instead of Clear(),
//                  otherwise we have a memory leak!
// 27 Dec 2000 (TG) end() optimization
// 03 Jan 2001 (TG) moved static variable storeRSEG_Flag to cluearte/cluearte.C
// 04 Jan 2001 (TG) added code for Delete(), changed Clear()
// 12 Jan 2001 (TG) removed obsolete header files
// 11 Feb 2001 (TG) #include RecoTrack.hh added
// 16 Mar 2001 (TG) added rad0 and hits in detector components
// 22 Mar 2001 (TG) added StreamUser()
// 28 Mar 2001 (TG) clear user array in Clear(), Delete()
// 08 Mai 2001 (TG) unlock(): added loop over RecoVertex with call to valid()
// 09 Jul 2001 (TG) unlock(): replaced unlock() calls by valid() call
// 11 Jul 2001 (TG) unlock(): Ecal->valid() added
// 14 Aug 2001 (TG) changed return type of print()
// 30 Nov 2001 (TG) improved StreamUser: added user.reserve(R__n)
// 02 Jan 2002 (TG) renamed prepareRecoTracksforVertexing() to
//                  PrepareRecoTracksForVertexing()
// 08 Jan 2002 (TG) unlock(): set all EcalHit to invalid() as default, because
//                  no track parameters are present at default.
//
// **********************************************************************
#include <time.h>
#include <iostream>
#include <iomanip>
#include <algorithm>

#include "clue/StreamObject.hh" // needed for RtEvent.hh
#include "gate/ObjArray.hh"     // needed for RtEvent.hh
#include "gate/ClonesArray.hh"  // needed for RtEvent.hh
#include "clue/cluearte.hh"     // included RtEvent.hh

#include "clue/Wire.hh"
#include "clue/EcalHit.hh"
#include "clue/MC_Track.hh"
#include "clue/MC_Vertex.hh"
#include "clue/RecoTrack.hh"
#include "clue/RecoVertex.hh"
#include "clue/TrackClones.hh"
#include "clue/CustomTrack.hh"


#if defined(USE_GATE)
#ifdef NAMESPACE
namespace CLUE {
#endif
  GClonesArray<RecoTrack*>*      RtEvent::recoTrackLink     = 0;
  GClonesArray<VDS_Track*>*      RtEvent::vdsTrackLink      = 0;
  GClonesArray<OTR_SpacePoint*>* RtEvent::otrSpacePointLink = 0;
  GClonesArray<RICH_Track*>*     RtEvent::richTrackLink     = 0;
  GClonesArray<EcalHit*>*        RtEvent::EcalHitLink       = 0;
  GClonesArray<MuonTrack*>*      RtEvent::muonTrackLink     = 0;
  GClonesArray<RecoVertex*>*     RtEvent::recoVertexLink    = 0;
  GClonesArray<TrackClones*>*    RtEvent::trackClonesLink   = 0;
  GClonesArray<MC_Vertex*>*      RtEvent::mcVertexLink      = 0;
  GClonesArray<MC_Track*>*       RtEvent::mcTrackLink       = 0;
  GClonesArray<Topp*>*           RtEvent::toppLink          = 0;
  GClonesArray<Wire*>*           RtEvent::wireLink          = 0;
#ifdef NAMESPACE
} // namespace CLUE
#endif

#endif // USE_GATE


#ifdef NAMESPACE
namespace CLUE {
#endif

  //==============================================================================
  // Constructors 
  //==============================================================================
  RtEvent::RtEvent() :
    Unixtime   ( 0 ),
    Evt        ( 0 ),
    Run        ( 0 ),
    Exp        ( 0 ),
    Vfld       ( 0.),
    Flagdst    ( 0 ),
    Dbmajor    ( 0 ),
    Dbminor    ( 0 ),
    Flag       ( 0 ),
    Bx         ( 0 ),
    Tcode      ( 0 ),
    Ir         ( 0.),
    Bg         ( 0.),
    Rad0       ( 0.),
    Hvds       ( 0 ),
    Hwch       ( 0 ),
    Hmsg       ( 0 ),
    Hptc       ( 0 ),
    Hric       ( 0 ),
    Htrd       ( 0 ),
    Hcal       ( 0 ),
    Hmuo       ( 0 ),
    Hmux       ( 0 ),
    Tracks     (new GClonesArray<RecoTrack*>( 100 ) ),
    Vds        (new GClonesArray<VDS_Track*>( 100 ) ),
    Otr        (new GClonesArray<OTR_SpacePoint*>( 100 ) ),
    Rich       (new GClonesArray<RICH_Track*>( 100 ) ),
    Ecal       (new GClonesArray<EcalHit*>( 100 ) ),
    Muon       (new GClonesArray<MuonTrack*>( 100 ) ),
    Vertices   (new GClonesArray<RecoVertex*>( 100 ) ),
    Wires      (new GClonesArray<Wire*>( 100 ) ),
    Clones     (new GClonesArray<TrackClones*>( 100 ) ),
    McTracks   (new GClonesArray<MC_Track*>( 100 ) ),
    McVertices (new GClonesArray<MC_Vertex*>( 100 ) ),
#ifdef USE_ROOT
    // MUST COME LAST!!
    streamer   (new StreamObject(this) )
#endif
    {
      if( toppLink==0 ) toppLink = new GClonesArray<Topp*>( 100 );
    }
  

  RtEvent::RtEvent( const RtEvent& rhs ) :
    Unixtime   ( rhs.Unixtime ),
    Evt        ( rhs.Evt ),
    Run        ( rhs.Run ),
    Exp        ( rhs.Exp ),
    Vfld       ( rhs.Vfld ),
    Flagdst    ( rhs.Flagdst ),
    Dbmajor    ( rhs.Dbmajor ),
    Dbminor    ( rhs.Dbminor ),
    Flag       ( rhs.Flag ),
    Bx         ( rhs.Bx ),
    Tcode      ( rhs.Tcode ),
    Ir         ( rhs.Ir ),
    Bg         ( rhs.Bg ),
    Rad0       ( rhs.Rad0 ),
    Hvds       ( rhs.Hvds ),
    Hwch       ( rhs.Hwch ),
    Hmsg       ( rhs.Hmsg ),
    Hptc       ( rhs.Hptc ),
    Hric       ( rhs.Hric ),
    Htrd       ( rhs.Htrd ),
    Hcal       ( rhs.Hcal ),
    Hmuo       ( rhs.Hmuo ),
    Hmux       ( rhs.Hmux ),
    Tracks     (new GClonesArray<RecoTrack*> ),
    Vds        (new GClonesArray<VDS_Track*> ),
    Otr        (new GClonesArray<OTR_SpacePoint*> ),
    Rich       (new GClonesArray<RICH_Track*> ),
    Ecal       (new GClonesArray<EcalHit*> ),
    Muon       (new GClonesArray<MuonTrack*> ),
    Vertices   (new GClonesArray<RecoVertex*> ),
    Wires      (new GClonesArray<Wire*> ),
    Clones     (new GClonesArray<TrackClones*> ),
    McTracks   (new GClonesArray<MC_Track*> ),
    McVertices (new GClonesArray<MC_Vertex*> ),
#ifdef USE_ROOT
    // MUST COME LAST!!
    streamer   (new StreamObject(this) )
#endif
    {
      cerr << "copy RtEvent NOT IMPLEMENTED!" << endl;
    }



  RtEvent::~RtEvent(){
#ifdef USE_ROOT
    TCollection::StartGarbageCollection();
    Delete();
    delete streamer;
    
    TCollection::EmptyGarbageCollection();
#else
    Clear();
#endif
    delete Tracks;
    delete Vds;
    delete Otr;
    delete Rich;
    delete Ecal;
    delete Muon;
    delete Vertices;
    delete Wires;
    delete Clones;
    delete McTracks;
    delete McVertices;
  }


  //==============================================================================
  // operator = 
  //==============================================================================
  RtEvent& RtEvent::operator=( const RtEvent& rhs ) {
    if( this == &rhs ) return *this;
    cerr << "RtEvent::operator= NOT IMPLEMENTED!" << endl;
    return *this;
  }


  //==============================================================================
  // Clear : empty containers, do not call DTOR of container elements
  //==============================================================================
#ifdef USE_ROOT
  void RtEvent::Clear(Option_t *)
#else
  void RtEvent::Clear()
#endif
    {
#ifdef USE_ROOT
      // Delete() calls DTOR of container elements, Clear() not!
      // -> therefore containers of classes with containers must
      //    use Delete(), otherwise we have a memory leak!
      Tracks->Clear(); //
      Vds->Clear();
      Otr->Clear();
      Rich->Clear();
      Ecal->Clear();
      Muon->Clear();
      Vertices->Clear(); //
      Wires->Clear();
      Clones->Clear();  //
      McTracks->Clear(); //
      McVertices->Clear();
#else
      Tracks->clear();
      Vds->clear();
      Otr->clear();
      Rich->clear();
      Ecal->clear();
      Muon->clear();
      Vertices->clear();
      Wires->clear();
      Clones->clear();
      McTracks->clear();
      McVertices->clear();
#endif
      user.clear();  // clear user array
    }


  //==============================================================================
  // Delete : call DTOR for array elements
  //==============================================================================
#ifdef USE_ROOT
  void RtEvent::Delete(Option_t *)
#else
  void RtEvent::Delete()
#endif
    {
#ifdef USE_ROOT
    Tracks->Delete();    
    Vds->Delete();       
    Otr->Delete();       
    Rich->Delete();      
    Ecal->Delete();      
    Muon->Delete();      
    Vertices->Delete();  
    Wires->Delete();     
    Clones->Delete();    
    McTracks->Delete();  
    McVertices->Delete();
#else
# warning missing code (T.Glebe, 04/01/2001)
#endif
    user.clear(); // clear user array
  }

  //==============================================================================
  // print
  //==============================================================================
  std::ostream& RtEvent::print( std::ostream& os ) const {
    char tstr[256];
    int length;
    const long int unixtim = unixtime();
    
    strcpy(tstr, asctime(localtime(&unixtim)));
    length = strlen(tstr)-1;
    if(tstr[length]=='\n') tstr[length] = '\0'; 
    
    os << "----->RtEvent<-----" << endl
       << "Experiment number: " << exp()      << endl
       << "Run Number:        " << run()      << endl
       << "Event:             " << evt()      << endl
       << "BX:                " << bx()       << endl
       << "DBmajor:           " << dbmajor()  << endl
       << "DBminor:           " << dbminor()  << endl
       << "Time:              " << unixtime() << " (" 
       << tstr << " UTC)" << endl
       << "Cherenkov radius:  " << rad0()   << endl
       << "Trigger code:      " << tcode()  << endl
       << "Interaction Rate:  " << ir()     << endl
       << "Background Rate:   " << bg()
       << endl;

    os << "Hits in VDS:           " << hvds()  << endl
       << "Hits in wire chambers: " << hwch()  << endl
       << "Hits in MSGC:          " << hmsg()  << endl
       << "Hits in PTC:           " << hptc()  << endl
       << "Hits in RICH:          " << hrich() << endl
       << "Hits in TRD:           " << htrd()  << endl
       << "Hits in ECAL:          " << hcal()  << endl
       << "Hits in outer Muon:    " << hmuo()  << endl
       << "Hits in inner Muon:    " << hmux()  << endl;

#ifdef USE_ROOT
    os << "Size of user field: " << user.size() << endl;
#endif

    os << "No. of RecoTracks:   " << setw(4) << Tracks->size()     << endl
       << "No. of RecoVertices: " << setw(4) << Vertices->size()   << endl
       << "No. of Clones:       " << setw(4) << Clones->size()     << endl
       << "No. of MCTracks:     " << setw(4) << McTracks->size()   << endl
       << "No. of MCVertices:   " << setw(4) << McVertices->size() << endl
       << "No. of wires:        " << setw(4) << Wires->size()      << endl 
       << "No. of VDS tracks:   " << setw(4) << Vds->size()        << endl
       << "No. of OTR tracks:   " << setw(4) << Otr->size()        << endl
       << "No. of RICH tracks:  " << setw(4) << Rich->size()       << endl
       << "No. of ECAL hits:    " << setw(4) << Ecal->size()       << endl
       << "No. of MUON tracks:  " << setw(4) << Muon->size()       << endl;
    
    return os << "-----------" << endl;
  }


  //==============================================================================
  // unlock 
  //==============================================================================
  void RtEvent::unlock() {
    const GClonesArray<RecoTrack*>::iterator rend = Tracks->end();
    for( GClonesArray<RecoTrack*>::iterator t = Tracks->begin();
	 t != rend; ++t ) { 
      (*t)->valid();
    }

    const GClonesArray<RecoVertex*>::iterator vtxend = Vertices->end();
    for( GClonesArray<RecoVertex*>::iterator t = Vertices->begin();
	 t != vtxend; ++t ) (*t)->set_valid();

    // EcalHit unlock
    const GClonesArray<EcalHit*>::iterator eend = Ecal->end();
    for( GClonesArray<EcalHit*>::iterator t = Ecal->begin();
	 t != eend; ++t ) (*t)->invalid();

    // RecoSegment unlock
    const GClonesArray<VDS_Track*>::iterator vend = Vds->end();
    for( GClonesArray<VDS_Track*>::iterator t = Vds->begin();
	 t != vend; ++t ) (*t)->valid();
    const GClonesArray<OTR_SpacePoint*>::iterator oend = Otr->end();
    for( GClonesArray<OTR_SpacePoint*>::iterator t = Otr->begin();
	 t != oend; ++t ) (*t)->valid();
    const GClonesArray<RICH_Track*>::iterator iend = Rich->end();
    for( GClonesArray<RICH_Track*>::iterator t = Rich->begin();
	 t != iend; ++t ) (*t)->valid();
    const GClonesArray<MuonTrack*>::iterator mend = Muon->end();
    for( GClonesArray<MuonTrack*>::iterator t = Muon->begin();
	 t != mend; ++t ) (*t)->valid();
  }


  //==============================================================================
  // PrepareRecoTracksForVertexing
  //==============================================================================
  bool RtEvent::PrepareRecoTracksForVertexing() {

    const GClonesArray<RecoTrack*>::iterator rend = Tracks->end();
    for( GClonesArray<RecoTrack*>::iterator t = Tracks->begin();
	 t != rend; ++t ) {
      RecoTrack* const r = *t;
      if(r->isValid())
	r->PrepareForVertexing();
    }

    return true;
  }

#ifdef USE_ROOT
  //==============================================================================
  // StreamUser
  //==============================================================================
  void RtEvent::StreamUser(TBuffer& R__b) {
    if (R__b.IsReading()) {
      user.clear();
      Int_t R__i, R__n;
      Float_t R__t;
      R__b >> R__n;
      //       cout << "x R__n: " << R__n << endl;
      user.reserve(R__n);
      for (R__i = 0; R__i < R__n; ++R__i) {
	R__b >> R__t;
	//      cout << " R__t: " << R__t << endl;
	user.push_back(R__t);
      }
    } else {
      R__b << static_cast<Int_t>(user.size());
      const vector<float>::iterator last = user.end();
      for (vector<float>::iterator R__k = user.begin(); R__k != last; ++R__k) {
	//      cout << " *R__k: " << *R__k << endl;
	R__b << *R__k;
      }
    }

    return;
  }

  //==============================================================================
  // Print 
  //==============================================================================
  void RtEvent::Print( Option_t * ) {
    print(std::cout);
  }

  
  //==============================================================================
  // Streamer 
  //==============================================================================
  void RtEvent::Streamer(TBuffer &) {
    cerr << "RtEvent::Streamer is not implemented!" << endl;
  }
  
  ClassImp(RtEvent)
#endif // USE_ROOT

#ifdef NAMESPACE
} // namespace CLUE
#endif
