#ifndef __GTOPP_HH
#define __GTOPP_HH
// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   23 Jan. 2001
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Replacement for ClPdtEntry
//
// changes:
// 23 Jan 2001 (TG) creation
// 24 Jan 2001 (TG) inheritance to ToppIf
// 26 Jan 2001 (TG) added with() member functions
// 29 Jan 2001 (TG) added anti_chrg(), anti_name()
// 28 Feb 2001 (TG) operator==() added
// 05 Mar 2001 (TG) moved anti() to anti_ptr(), added new anti()
// 12 Mar 2001 (TG) anychrg, CTOR added
// 27 Apr 2001 (TG) swith()
// 28 Jun 2001 (TG) removed template parameters C and T
// 03 Jul 2001 (TG) createPattern() added
// 15 Aug 2001 (TG) changed return type of topp()
// 17 Aug 2001 (TG) moved inline functions to .icc file
// 05 Sep 2001 (TG) added lund()
// 14 Sep 2001 (TG) renamed sdecaysTo() to decaysTo() and sswith() to with()
// 19 Dec 2001 (TG) protected P_trait.hh against CINT
//
// ********************************************************************
#include <iosfwd>
#include "arte/TObject.hh"
#include "interfaces/ToppIf.hh"
#ifndef __CINT__
#include "pattern/P_trait.hh"
#endif

template <class T> class ArtePointer;
template <class C, class T> class SProperty;
class Topp;
class CutIf;
class RecoTrack;

/** Representation of a Topp table entry.

    This class represents a Topp table entry via a pointer to the
    ArteTable<Topp>. In addition, factory functions are provided to generate
    objects which are needed for a pattern based analysis.

    @author T. Glebe
 */
//==============================================================================
// GTopp
//==============================================================================
class GTopp : public ToppIf {
public:
  /** @name --- Constructors --- */
  ///
  GTopp();
  ///
  GTopp( const GTopp& rhs );
  ///
  GTopp( const char* name );
  ///
  GTopp( const char* name, int chrg );
  ///
  virtual ~GTopp();

  /** @name --- ToppIf access methods --- */
  /// particle name
  const char* const name() const;
  /// mass
  float mass() const;
  /// width
  float wid()  const;
  /// truncate Breit-Wigner at mass - trbw
  float trbw() const;
  /// life time
  float life() const;
  /// 3 * charge
  int   chrg() const;
  /// access to anti-particle
  const ToppIf& anti() const;
  /// LUND code
  int lund()   const;

  /** @name --- Access methods --- */
  /// topp table pointer
  const ToppIf& topp() const;
  /// Do we represent a valid Topp entry?
  bool valid() const;

#ifndef __CINT__
  /** @name --- Factory functions --- */
  /// create a SPattern object
  SP_sp swith(const SProperty<TrackIf,bool>& rhs) const;
  ///
  PMC_sp createPattern() const;
  ///
  PMC_sp with(const SProperty<TrackIf,bool>& rhs) const;

  /// create a SDecayPatternMatchContainer object
  template <unsigned int NTR>
  typename DP_trait<NTR>::DPMC_sp
  decaysTo(const SPatternExpression<NTR>& rhs) const;
#endif

  /** @name --- Operators --- */
  ///
  bool operator==(const ToppIf& rhs) const;
  ///
  bool operator==(const GTopp& rhs) const;
  ///
  bool operator!=(const GTopp& rhs) const;
  ///
  bool operator< (const GTopp& rhs) const;

  /** @name --- Expert functions --- */
  /// set topp table pointer
  void set_topp(const ArtePointer<Topp>& topp);
  /// set link to anti GTopp object
  void set_anti(const GTopp& anti);
  
  /// used by operator<<()
  std::ostream& print( std::ostream& os ) const;
    
private:
  UShort_t       iTopp;                             // Topp table id
  const ToppIf*  anti_;
  int            anychrg;                           // a charge in case of Any
}; // end of GTopp


//==============================================================================
// operator <<
//==============================================================================
std::ostream& operator<<( std::ostream& os, const GTopp& a );

#ifndef __CINT__
#include "clue/GTopp.icc"
#endif
#endif
