// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   22 Mai 2000
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: 
//
// changes:
// 22 Mai 2000 (TG) creation
// 20 Nov 2000 (TG) added namespace CLUE
// 12 Jan 2001 (TG) added WireDirection(), Propagate()
// 24 Jan 2001 (TG) removed obsolete include directives
//
// *****************************************************************************
#include <iostream>
#include <cmath>
#include "geometry/Point3D.hh"  // needed by RecoSegment.hh, ClueUtil.hh
#include "clue/RecoSegment.hh"
#include "clue/Wire.hh"
#include "clue/ClueUtil.hh"
#include "clue/MC_Track.hh"

#ifdef NAMESPACE
namespace CLUE {
#endif

  //==============================================================================
  // WireDirection: return unit Vector of Wire direction
  //==============================================================================
  const ThreeVector WireDirection( const Wire& rhs ) {
    return (rhs.dx() > rhs.dy()) ? ThreeVector(1,0,0) : ThreeVector(0,1,0);
  }


  //==============================================================================
  // Propagate: calculate impact point at z
  //==============================================================================
  const Point3D Propagate(const RecoSegment& rs, float z) {
    const float dz = z - rs.z();
    return Point3D( rs.x() + rs.tx() * dz, rs.y() + rs.ty() * dz, z); 
  }


  //==============================================================================
  // distance - between two rays
  //==============================================================================
  const double distance(const ThreeVector& r1, const ThreeVector& m1,
			const ThreeVector& r2, const ThreeVector& m2) {
    const ThreeVector dr = r1 - r2;
    const double denom = m1.Cross(m2).Mag();
    
    const double detx = dr.x() * (m1.y() * m2.z() - m1.z() * m2.y());
    const double dety = dr.y() * (m1.z() * m2.x() - m1.x() * m2.z());
    const double detz = dr.z() * (m1.x() * m2.y() - m1.y() * m2.x());
    
    return fabs(detx + dety + detz) / denom;
  }


  //==============================================================================
  // distance - between RecoSegment & Wire
  //==============================================================================
  const double distance( const RecoSegment& rseg, const Wire& wire) {
    
    const ThreeVector r1(rseg.x(), rseg.y(), rseg.z());
    const ThreeVector m1(rseg.tx(), rseg.ty(), 1.);
    const ThreeVector r2(wire.x(), wire.y(), wire.z());
    const ThreeVector m2(WireDirection(wire));
    
    return distance(r1,m1,r2,m2);
  }

  const double distance( const Wire& wire, const RecoSegment& rseg) {
    return distance(rseg, wire); 
  }

#ifdef NAMESPACE
} // namespace CLUE
#endif
