// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   20. Jul 1997
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Class to describe a particle decay pattern
//
// changes:
// 20 Jul 1997 creation by C. Borgmeier, I. Siccama
// 03 Nov 2000 (TG) removed #include "arte/arte.hh", added a constructor
// 17 Nov 2000 (TG) modified header files
// 20 Nov 2000 (TG) added namespace CLUE
// 27 Dec 2000 (TG) end() optimization
// 12 Jan 2001 (TG) removed obsolete include files
//
// *****************************************************************************
#include <string>
#include <iostream>
#include <algorithm>
#include <map>

#include "clue/cluearte.hh"
// #include "geometry/Point3D.hh"   // needed by RecoSegment.hh
#include "clue/SmartPointer.hh"  // needed by Pattern.hh
#include "gate/ObjArray.hh"      // needed for Pattern.hh
#include "gate/ClonesArray.hh"   // needed for Pattern.hh
#include "clue/Pattern.hh"
#include "clue/Topp.hh"
#include "clue/RecoVertex.hh"
// #include "geometry/ClLorentzVector.hh" // needed by MC_Track.hh
#include "clue/MC_Track.hh"
#include "clue/PdtEntry.hh"

#ifdef NAMESPACE
namespace CLUE {
#endif

  // NULL pointer
  static const ArtePointer<Topp> topp0;

  using  std::string;  
  static std::multimap<string,ClPdtEntry*>* pdtMap = 0;


  //==============================================================================
  // updateTOPP
  //==============================================================================
  // this function sets the ArtePointer<Topp> of the global ClPdtEntry objects.
  void updateTOPP2() {
    const ArteTable<Topp>::iterator end = ArteTable<Topp>::end();
    for( ArteTable<Topp>::iterator tit = ArteTable<Topp>::begin();
	 tit != end; ++tit ) {
      const ArtePointer<Topp> topp = *tit;

      string searchName = string( topp->name() );
      searchName.erase( std::find( searchName.begin(), searchName.end(), ' ' ),
			searchName.end() );
      
      std::pair<std::multimap<string,ClPdtEntry*>::iterator,
	std::multimap<string,ClPdtEntry*>::iterator> range
	= pdtMap->equal_range( searchName );
      
      const std::multimap<string,ClPdtEntry*>::iterator mend = pdtMap->end();
      for( std::multimap<string,ClPdtEntry*>::iterator i = range.first;
	   i!=mend && searchName == (*i).second->name; ++i ) {
	//      cout << "*i.second->name: " << (*i).second->name << " *topp: " 
	//	   << *topp << endl;
	(*i).second->set_topp( topp );
      }
    }
  }


  //==============================================================================
  // Constructors
  //==============================================================================
  ClPdtEntry::ClPdtEntry(): iTopp( 0 ) {
    name[0] = 0;
  }


  ClPdtEntry::ClPdtEntry( const char* n ): iTopp( 0 ) {
    //cerr<<"Creating Entry: "<<index()<<endl;
    strncpy( name, n, 12 );
  
    if( pdtMap==0 ) pdtMap = new std::multimap<string,ClPdtEntry*>;
    pdtMap->insert( std::make_pair( string( name ), this ) );
  }


  ClPdtEntry::ClPdtEntry( const char* n,
			  const ArtePointer<Topp>& t ) :
    iTopp( getIDX<Topp>( t ) ) {
    //cerr<<"Creating Entry: "<<index()<<endl;
    strncpy( name, n, 12 );
    
    if( pdtMap==0 ) pdtMap = new std::multimap<string,ClPdtEntry*>;
    pdtMap->insert( std::make_pair( string( name ), this ) );
  }


  ClPdtEntry::ClPdtEntry( const char* n, unsigned int t ) :
    iTopp( t ) {
    //  : iTopp( *(ArteTable<Topp>::begin() + t - 1) ) {
    //cerr<<"Creating Entry: "<<index()<<endl;
    strncpy( name, n, 12 );
    
    if( pdtMap==0 ) pdtMap = new std::multimap<string,ClPdtEntry*>;
    pdtMap->insert( std::make_pair( string( name ), this ) );
  }


  ClPdtEntry::~ClPdtEntry() {
    //cerr << "deleting entry" << endl;
    const string n( name );
    std::multimap<string,ClPdtEntry*>::iterator upper = pdtMap->upper_bound( n );
    std::multimap<string,ClPdtEntry*>::iterator result
      = std::find( pdtMap->lower_bound( n ), upper,
		   std::pair<const string, ClPdtEntry*>( n, this ) );
    if( result!=upper ) pdtMap->erase( result );
  }


  //==============================================================================
  // operator =
  //==============================================================================
  const ClPdtEntry& ClPdtEntry::operator=( const ClPdtEntry& rhs ) {
    if( this == &rhs ) return *this;
    cerr << "not implemented: ClPdtEntry::operator=" << endl;
    return *this;
  }


  //==============================================================================
  // mass
  //==============================================================================
  double ClPdtEntry::mass() const {
    if( topp() != topp0 ) return topp()->mass();
    else return 0;
  }


  //==============================================================================
  // print
  //==============================================================================
  void ClPdtEntry::print( std::ostream& os ) const {
    //os << topp;
    if( topp() != topp0 ) os << *topp();
    else os << name;
  }


  //==============================================================================
  // operator== not inline, to avoid strong coupling with other classes
  //==============================================================================
  bool ClPdtEntry::operator==( const ArtePointer<Topp>& rhs ) const 
  { return topp() == rhs; }

  bool ClPdtEntry::operator!=( const ArtePointer<Topp>& rhs ) const
  { return topp() != rhs; }

  bool ClPdtEntry::operator==( const RecoTrack& rhs ) const 
  { return topp() == rhs.topp(); }

  bool ClPdtEntry::operator!=( const RecoTrack& rhs ) const
  { return topp() != rhs.topp(); }

  bool ClPdtEntry::operator==( const RecoVertex& rhs ) const
  { return topp() == rhs.topp(); }

  bool ClPdtEntry::operator!=( const RecoVertex& rhs ) const
  { return topp() != rhs.topp(); }

  bool ClPdtEntry::operator==( const MC_Track& rhs ) const
  { return topp() == rhs.topp(); }

  bool ClPdtEntry::operator!=( const MC_Track& rhs ) const
  { return topp() != rhs.topp(); }


  //==============================================================================
  // decaysTo
  //==============================================================================
  Pattern ClPdtEntry::decaysTo( Pattern const& l ) const {
    return Pattern( *this, l );
  }


  //==============================================================================
  // operator &
  //==============================================================================
  Pattern ClPdtEntry::operator&( ClPdtEntry const& e ) const {
    //cerr << "Entry: " << *this << " anding entry:" << e << endl;
    return Pattern( *this ) & e;
  }


  //==============================================================================
  // operator |
  //==============================================================================
  Pattern ClPdtEntry::operator|( ClPdtEntry const& e ) const {
    //cerr << "Entry: " << *this << " oring entry:" << e << endl;
    return Pattern( *this ) | e;
  }


  //==============================================================================
  // operator <
  //==============================================================================
  bool ClPdtEntry::operator<( ClPdtEntry const& e ) const {
    return topp() < e.topp();
  }


  //==============================================================================
  // operator &
  //==============================================================================
  Pattern ClPdtEntry::operator&( Pattern const& e ) const {
    //cerr << "Entry: " << *this << " anding node:" << e << endl;
    return Pattern( *this ) & e;
  }


  //==============================================================================
  // operator |
  //==============================================================================
  Pattern ClPdtEntry::operator|( Pattern const& e ) const {
    //cerr << "Entry: " << *this << " oring node:" << e << endl;
    return Pattern( *this ) | e;
  }


  //==============================================================================
  // operator ()
  //==============================================================================
  Pattern ClPdtEntry::operator()( const ClProperty<RecoTrack,int>& prop ) const {
    return Pattern( *this, &prop );
  }
  
  
  //==============================================================================
  // andp
  //==============================================================================
  Pattern ClPdtEntry::andp( ClPdtEntry const& e ) const { 
    return *this & e;
  }


  //==============================================================================
  // orp
  //==============================================================================
  Pattern ClPdtEntry::orp( ClPdtEntry const& e ) const {
    return *this | e;
  }


  //==============================================================================
  // andp
  //==============================================================================
  Pattern ClPdtEntry::andp( Pattern const& e ) const {
    return *this & e;
  }


  //==============================================================================
  // orp
  //==============================================================================
  Pattern ClPdtEntry::orp( Pattern const& e ) const {
    return *this | e;
  }


  //==============================================================================
  // with
  //==============================================================================
  Pattern ClPdtEntry::with( const ClProperty<RecoTrack,int>& prop ) const {
    return (*this)( prop );
  }
  
  //==============================================================================
  // select
  //==============================================================================
  int ClPdtEntry::select( const GObjArray<RecoTrack*>& tracks,
			  GObjArray<PatternMatch*>& result ) const {
    const GObjArray<TrackBase*>& casted
      = reinterpret_cast<const GObjArray<TrackBase*>&>( tracks );
    return Pattern(*this).select( casted, result );
  }


  int ClPdtEntry::select( const GObjArray<MC_Track*>& tracks,
			  GObjArray<PatternMatch*>& result ) const {
    return Pattern(*this).select( tracks, result );
  }

#ifdef USE_ROOT
  int ClPdtEntry::select( const GClonesArray<RecoTrack*>& tracks,
			  GObjArray<PatternMatch*>& result ) const {
    const GClonesArray<TrackBase*>& casted
      = reinterpret_cast<const GClonesArray<TrackBase*>&>( tracks );
    return Pattern(*this).select( casted, result );
  }
  
  
  int ClPdtEntry::select( const GClonesArray<MC_Track*>& tracks,
			  GObjArray<PatternMatch*>& result ) const {
    return Pattern(*this).select( tracks, result );
  }
#endif

#ifdef USE_ROOT
  //==============================================================================
  // Print
  //==============================================================================
  void ClPdtEntry::Print( Option_t* ) {
    print( std::cout );
    std::cout << endl;
  }

  //==============================================================================
  // Streamer
  //==============================================================================
  void ClPdtEntry::Streamer(TBuffer &) {
    cerr << "ClPdtEntry::Streamer not implemented!" << endl;
  }
  
  ClassImp(ClPdtEntry)
#endif

  ClPdtEntry ANY( "Any" );
  ClPdtEntry X( "X" );

#ifdef NAMESPACE
} // namespace CLUE
#endif
