#ifndef PROPERTY_PATTERN_HH
#define PROPERTY_PATTERN_HH
// **********************************************************************
//
// source:
//
// type:      source code
//
// created:   16. Sep 1998
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Class describing a member function and a search pattern
//
// changes:
// 16 Sep 1998 creation by C. Borgmeier
// 03 Nov 2000 (TG) code polishing
// 14 Nov 2000 (TG) inlined some members of ClProperty
// 20 Nov 2000 (TG) added namespace CLUE
// 24 Nov 2000 (TG) added lt() member function as operator< has problems in CINT
//
// **********************************************************************

#include <iosfwd>
#include <iostream>
#include "arte/TObject.hh"
#include "gate/RtypesT.h"
#include "gate/gatefwd.hh"
#include "clue/cluefwd.hh"

#ifdef NAMESPACE
namespace CLUE {
#endif

  template <class T> int lessfunc(T lhs, T rhs);
  template <class T> int morefunc(T lhs, T rhs);


  /** Expression template base class.
      @memo class ClProperty
      @author C. Borgmeier
  */
  //==============================================================================
  // ClProperty
  //==============================================================================
  template <class C, class T>
    class ClProperty: public TObject {
    public:
      ///
      ClProperty() {}
      ///
      virtual ~ClProperty();
      // This operator tests the property
      virtual T operator()( const C& ) const;
      // This operator tests the property
      virtual T operator()( const TrackBase& ) const;
      ///
      virtual ClProperty<C,T>* propertyClone() const;
      
      /// less than - operator< has problems in CINT
      inline ClExpression<C,T,int> lt( const ClProperty<C,T>& rhs ) const {
	return ClExpression<C,T,int>( *this, rhs, &lessfunc );
      }
      /// less than - operator< has problems in CINT
      inline ClExpression<C,T,int> lt( T rhs ) const {
	//    return (*this < ClConstant<C,T>( rhs ));
	return operator<( ClConstant<C,T>( rhs ) );
      }

      ///
      inline ClExpression<C,T,int> operator<( const ClProperty<C,T>& rhs ) const {
	return ClExpression<C,T,int>( *this, rhs, &lessfunc );
      }
      ///
      inline ClExpression<C,T,int> operator<( T rhs ) const {
	//    return (*this < ClConstant<C,T>( rhs ));
	return operator<( ClConstant<C,T>( rhs ) );
      }
      ///
      inline ClExpression<C,T,int> operator>( const ClProperty<C,T>& rhs ) const {
	return ClExpression<C,T,int>( *this, rhs, &morefunc );
      }
      ///
      ClExpression<C,T,int> operator>( T rhs ) const {
	return operator>( ClConstant<C,T>( rhs ) );
      }
      ///
      ClExpression<C,T,int> operator==( const ClProperty<C,T>& rhs ) const;
      ///
      ClExpression<C,T,int> operator==( T rhs ) const;
      
      ///
      ClExpression<C,T,int> operator!=( const ClProperty<C,T>& rhs ) const;
      ///
      ClExpression<C,T,int> operator!=( T rhs ) const;
      
      ///
      ClExpression<C,T,T> operator+( const ClProperty<C,T>& rhs ) const;
      ///
      ClExpression<C,T,T> operator+( T rhs ) const;
      
      ///
      ClExpression<C,T,T> operator-( const ClProperty<C,T>& rhs ) const;
      ///
      ClExpression<C,T,T> operator-( T rhs ) const;
      
      ///
      ClExpression<C,T,T> operator*( const ClProperty<C,T>& rhs ) const;
      ///
      ClExpression<C,T,T> operator*( T rhs ) const;
      
      ///
      ClExpression<C,T,T> operator/( const ClProperty<C,T>& rhs ) const;
      ///
      ClExpression<C,T,T> operator/( T rhs ) const;
      
      ///
      ClExpression<C,T,int> operator&( const ClProperty<C,T>& rhs ) const;
      ///
      ClExpression<C,T,int> operator|( const ClProperty<C,T>& rhs ) const;
      
      ///
      GObjArray<C*> select( const GObjArray<C*>& from ) const;
      ///
      int select( const GObjArray<C*>& from, GObjArray<C*>& to ) const;
      
#ifdef USE_ROOT
      ///
      GObjArray<C*> select( const GClonesArray<C*>& from ) const;
      ///
      int select( const GClonesArray<C*>& from, GObjArray<C*>& to ) const;
#endif

      ///
      GObjArray<C*> selectremove( GObjArray<C*>& from ) const;
      
      /// used by operator<<()
      virtual void print( std::ostream& ) const;

      // replacement for normal ==
      //virtual bool isEqualProperty( const ClProperty& rhs ) const; 
    protected:
      ClProperty( const ClProperty<C,T>& rhs ) { operator=( rhs );   }
      const ClProperty<C,T>& operator=( const ClProperty<C,T>& );
      
    private:
      void pushTrack( const TrackBase& track,
		      GObjArray<PatternMatch*>& pList ) const;
      
#ifdef USE_ROOT
      ClMyDefT2(ClProperty<C,T>,1);
#endif
    }; // end of class ClProperty


  // template <class C, class T>
  // std::ostream& operator<<( std::ostream& os, const ClProperty<C,T>& prop );


  /** Expression template.
      @memo class ClExpression
      @author C. Borgmeier
  */
  //==============================================================================
  // ClExpression
  //==============================================================================
  template <class C, class T, class X>
    class ClExpression: public ClProperty<C,X> {
    public:
      ///
      ClExpression();
      ///
      ClExpression( const ClProperty<C,T>& lhs,
		    const ClProperty<C,T>& rhs,
		    X (*func)( T, T ) );
      ///
      ClExpression( const ClExpression<C,T,X>& rhs ) { operator=( rhs ); }
      ///
      ~ClExpression();
      ///
      const ClExpression<C,T,X>& operator=( const ClExpression<C,T,X>& );
      ///
      virtual ClProperty<C,X>* propertyClone() const;
      //
      X operator()( const C& obj ) const;
      // This operator tests the property
      X operator()( const TrackBase& ) const;
      
      /// used by operator<<()
      virtual void print( std::ostream& ) const;
      
    private:
      typedef ClProperty<C,T> P_type; // to help CINT parsing

      P_type* lhs;
      P_type* rhs;
      X (*func)( T, T );
      
#ifdef USE_ROOT
      ClMyDefT3(ClExpression<C,T,X>, 1);
#endif
  }; // ClExpression


  /** class ClMember
      @author C. Borgmeier
  */
  //==============================================================================
  // ClMember
  //==============================================================================
  template <class C, class T>
    class ClMember: public ClProperty<C,T> {
    public:
      ///
      ClMember() {}
      ///
      ~ClMember();
      
      ///
      virtual ClProperty<C,T>* propertyClone() const;
      
      /// used by operator<<()
      virtual void print( std::ostream& ) const;
      
#ifndef CINTOBJECT
      ClMember( T (C::*f)() const );
#endif
      
      //
      T operator()( const C& part ) const;
      //
      T operator()( const TrackBase& part ) const;
      
    private:
      T (C::*func)() const;
      
#ifdef USE_ROOT
      ClMyDefT2(ClMember<C,T>,1);
#endif
  }; // ClMember


  /** Unary expression template.
      @memo class ClMemberOnePar
      @author C. Borgmeier
  */
  //==============================================================================
  // ClMemberOnePar
  //==============================================================================
  template <class C, class T, class P>
    class ClMemberOnePar: public ClProperty<C,T> {
    public:
      ///
      ClMemberOnePar() {}
      ///
      ~ClMemberOnePar();
      
      ///
      virtual ClProperty<C,T>* propertyClone() const;
      
      /// used by operator<<()
      virtual void print( std::ostream& ) const;
      
#ifndef CINTOBJECT
      ClMemberOnePar( T (C::*f)( P ) const, P par );
#endif
      
      //
      T operator()( const C& part ) const;
      //
      T operator()( const TrackBase& part ) const;

    private:
      T (C::*func)( P par ) const;
      P par;
      
#ifdef USE_ROOT
      ClMyDefT3(ClMemberOnePar<C,T,P>,1);
#endif
  }; // ClMemberOnePar


  /** Binary expression template.
      @memo ClMemberTwoPar
      @author C. Borgmeier
  */
  //==============================================================================
  // ClMemberTwoPar
  //==============================================================================
  template <class C, class T, class P1, class P2>
    class ClMemberTwoPar: public ClProperty<C,T> {
    public:
      ///
      ClMemberTwoPar() {}
      ///
      ~ClMemberTwoPar();
      
      ///
      virtual ClProperty<C,T>* propertyClone() const;

      /// used by operator<<()
      virtual void print( std::ostream& ) const;

#ifndef CINTOBJECT
      ClMemberTwoPar( T (C::*f)( P1, P2 ) const, P1 par1, P2 par2 );
#endif
  
      //
      T operator()( const C& part ) const;
      //
      T operator()( const TrackBase& part ) const;
      
    private:
      T (C::*func)( P1 par1, P2 par2 ) const;
      P1 par1;
      P2 par2;
      
#ifdef USE_ROOT
      ClMyDefT4(ClMemberTwoPar<C,T,P1,P2>,1);
#endif
  }; // ClMemberTwoPar


  /** Constant expression template.
      @memo ClConstant
      @author C. Borgmeier
  */
  //==============================================================================
  // ClConstant
  //==============================================================================
  template <class C, class T>
    class ClConstant: public ClProperty<C,T> {
    public:
      ///
      ClConstant() {}
      ///
      ClConstant( T constant );
      ///
      ~ClConstant();
      
      ///
      virtual ClProperty<C,T>* propertyClone() const;
      
      //
      T operator()( const C& part ) const;
      //
      T operator()( const TrackBase& part ) const;
      
      /// used by operator<<()
      virtual void print( std::ostream& ) const;
      
    private:
      T constant;
  
#ifdef USE_ROOT
      ClMyDefT2(ClConstant<C,T>,1);
#endif
    }; // class ClConstant
  

#ifdef NAMESPACE
} // namespace CLUE

#ifdef USE_ROOT
MyClMyInstT2(CLUE,ClProperty,C,T);
MyClMyInstT3(CLUE,ClExpression,C,T,X);
MyClMyInstT2(CLUE,ClMember,C,T);
MyClMyInstT3(CLUE,ClMemberOnePar,C,T,P);
MyClMyInstT4(CLUE,ClMemberTwoPar,C,T,P1,P2);
MyClMyInstT2(CLUE,ClConstant,C,T);
#endif
#else  // NAMESPACE
#ifdef USE_ROOT
ClMyInstT2(ClProperty,C,T);
ClMyInstT3(ClExpression,C,T,X);
ClMyInstT2(ClMember,C,T);
ClMyInstT3(ClMemberOnePar,C,T,P);
ClMyInstT4(ClMemberTwoPar,C,T,P1,P2);
ClMyInstT2(ClConstant,C,T);
#endif
#endif // NAMESPACE


#endif // PROPERTY_PATTERN_HH
