// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   Die Okt  2 12:18:59 MEST 2001
//
// author:    glebe
//
// Description: 
//
// changes:
//
// ********************************************************************
#include <iostream>
#include "TH1.h"
#include "TH2.h"
#include "TNtuple.h"
#include "TPaveText.h"
#include "BEE.hh"
#include "include/v0Analysis.hh"

//==============================================================================
// Constructor
//==============================================================================
v0Analysis::v0Analysis(RtEvent* evt) :
 event  (evt),
 nacc   (0),
 nrej   (0),
 pidyes (0),
 pidno  (0),
 Piplus (false),
 Piminus(false),
 Kplus  (false),
 Kminus (false),
 Muplus (false),
 Muminus(false),
 Kshort (false),
 Phi    (false),
 JPsi   (false),
 B0     (false)
{}

//==============================================================================
// EventSelect
//==============================================================================
bool v0Analysis::EventSelect() {

  // select only events with 1 primary vertex
  if(ArteTable<RecoVertex>::size() != 1) {
    ++nrej;
    return false;
  }

  ++nacc;
  return true;
}

//==============================================================================
// DefinePattern
//==============================================================================
bool v0Analysis::DefinePattern() {

  Piminus    = pi_minus.createPattern();
  Piplus     = pi_plus.createPattern();
  Kminus     = K_minus.createPattern();
  Kplus      = K_plus.createPattern();
  Muminus    = mu_minus.createPattern();
  Muplus     = mu_plus.createPattern();

  Kshort = K_S.decaysTo(Piminus & Piplus);
  Phi    = phi.decaysTo(Kminus & Kplus);
  JPsi   = Jpsi.decaysTo(Muminus & Muplus);
  B0     = B_zero.decaysTo(Piminus & Piplus);
  B0b    = B_zero.decaysTo(Piminus & Kplus);
  B0bar  = B_zero.decaysTo(Kminus & Piplus);

  JPsi->massrange(0.5);
  B0->massrange(1.);
  B0b->massrange(1.);
  B0bar->massrange(1.);

  return true;
}

//==============================================================================
// DefineHistograms
//==============================================================================
bool v0Analysis::DefineHistograms() {

  kshort = new TNtuple("kshort","K0s","m:dmin:dist:ctau:ctpt:trig");
  ptxf   = new TH2F("ptxf", "pt vs. xf(pion)", 100, 0., 1., 100, -0.4, 0.4);
  mom    = new TH1F("mom", "Momentum", 100, 0., 50.);

  Nphi   = new TNtuple("Nphi","Phi","m:dmin:dist:ctau:ctpt");
  Njpsi  = new TNtuple("Njpsi","Jpsi","m:dmin:dist:ctau:ctpt:rms");
  NB     = new TNtuple("NB","B0","m:dmin:dist:ctau:ctpt:run:pt1:pt2:trig:rms:evt:p1:p2:pt:p:pid:ang:dx:dz:tp:tpt:tpid:ttrig:ntri:mkpi:mpik:htx:hty:kpm");
  NBkp   = new TNtuple("NBkp", "B0->Kpi", "m:dmin:dist:ctau:ctpt:run:pt1:pt2:trig:rms:p1:p2:pt:p:pid");

  tri    = new TH1F("tri", "triggerbits/mass", 40, 5.0, 5.4);

  return true;
}

//==============================================================================
// Event
//==============================================================================
bool v0Analysis::Event() {

  // clear Particle container
  Piplus->clear();
  Piminus->clear();
  Kplus->clear();
  Kminus->clear();
  Muplus->clear();
  Muminus->clear();

  //----------------- select Particles -------------------
  const ArteTableIterator<RecoTrack> RTend = ArteTable<RecoTrack>::end();
  for(ArteTableIterator<RecoTrack> RTit = ArteTable<RecoTrack>::begin();
      RTit != RTend; ++RTit) {
    RecoTrack& rtra = **RTit;
  
    const unsigned int PID = rtra.pid();

    // PID statistics
    if(PID != 0) {
      ++pidyes;
    } else
      ++pidno;

    ptxf->Fill(rtra.pt(), rtra.xf(pi_plus.mass()));
    mom->Fill(rtra.p());
    if(rtra.pt()<0.1 || rtra.p() < 2. || fabs(rtra.xf(pi_plus.mass())) > 0.2)
      continue;

    // prepare track for vertexing
    //    rtra.PrepareForVertexing();
    const int chrg = rtra.charge();

    switch(PID) {
    case PID::kaon:
      if(chrg > 0) {
	Kplus->push_back(rtra);
	//      cout << " K+ found " << endl;
      }
      if(chrg < 0) {
	Kminus->push_back(rtra);
	//      cout << " K- found " << endl;
      }
      break;
    case PID::muon:
      if(chrg > 0) {
	Muplus->push_back(rtra);
	//      cout << " Mu+ found " << endl;
      }
      if(chrg < 0) {
	Muminus->push_back(rtra);
	//      cout << " Mu- found " << endl;
      }
      break;
    case 0:
      // WARNING: FALLTHROUGH!
    case PID::pion:
      if(chrg > 0) {
	Piplus->push_back(rtra);
	//      cout << " pi+ found " << endl;
      }
      if(chrg < 0) {
	Piminus->push_back(rtra);
	//      cout << " pi- found " << endl;
      }
      break;
    } // switch
    
  } // for RecoTrack


  //------------------ Pattern ----------------------
  Kshort->combine_no_vertex();
  Phi->combine_no_vertex();
  JPsi->combine_no_vertex();
  B0->combine_no_vertex();
  B0b->combine_no_vertex();
  B0bar->combine_no_vertex();

  // get primary vertex
  const RecoVertex& primvtx = *ArtePointer<RecoVertex>(1);

  //------------ Loop over Kshort Candidates -----------
  const TwoProngCandidateIterator KPM_end = Kshort->matches().end();
  for(TwoProngCandidateIterator PM_it = Kshort->matches().begin();
      PM_it != KPM_end; ++PM_it) {
    TwoProngCandidate k0cand = *PM_it;

    SVertex<2>& vtx = *k0cand->vertex();
    SVector<double,3> pos;
    vtx.findVertex3D(pos); // compute vertex manually
    vtx.set_vpos(pos);     // set vertex position manually
    SVector<double,2> rd = vtx.rmsDistAngle(); // calc rms distance
    // cut on rms distance
    if(rd[0] > 0.05) { k0cand->invalid(); continue; }
    k0cand->calc_mother_tr();  // must be calculated after vpos is set!

    double dmin = TVdistance(*k0cand, primvtx);     // track-vtx distance
    double dist = SGNVVdistance(*k0cand, primvtx);  // signed vtx-vtx distance

    if(dist < -0.1) { k0cand->invalid(); continue; }

    double ctau = k0cand->mass() * dist / k0cand->p();

    const SVector<double,3> PtotUnit(vtx.pvec().unit());
    const SVector<double,3> P1(vtx.track(1)->pvec());

    double P1_l = dot(P1, PtotUnit);
    // double frac     = (Pplus_l - Pminus_l) / (Pplus_l + Pminus_l);
    double pt = mag(P1 - PtotUnit * P1_l);

    const PatternMatchIf& t1 = *k0cand->PatternMatch(0);
    const PatternMatchIf& t2 = *k0cand->PatternMatch(1);
    const RecoTrack& r1 = t1.getRecoTrack();
    const RecoTrack& r2 = t2.getRecoTrack();
    unsigned int trig = 0;
    if( r1.cmp().slte() || r1.cmp().sltmu() ) ++trig;
    if( r2.cmp().slte() || r2.cmp().sltmu() ) ++trig;

    float x[6];
    x[0] = k0cand->mass();
    x[1] = dmin;
    x[2] = dist;
    x[3] = ctau;
    x[4] = ctau*pt;
    x[5] = trig;

    kshort->Fill(x);
  } // for SDPM

  //------------ Loop over Phi Candidates -----------
  const TwoProngCandidateIterator PPM_end = Phi->matches().end();
  for(TwoProngCandidateIterator PM_it = Phi->matches().begin();
      PM_it != PPM_end; ++PM_it) {
    TwoProngCandidate phicand = *PM_it;

    SVertex<2>& vtx = *phicand->vertex();
    SVector<double,3> pos;
    vtx.findVertex3D(pos);      // compute vertex manually
    vtx.set_vpos(pos);          // set vertex position manually
    phicand->calc_mother_tr();  // must be calculated after vpos is set!

    double dmin = TVdistance(*phicand, primvtx);     // track-vtx distance
    double dist = SGNVVdistance(*phicand, primvtx);  // signed vtx-vtx distance

    //    if(dist < -0.1) { phicand->invalid(); continue; }

    double ctau = phicand->mass() * dist / phicand->p();

    const SVector<double,3> PtotUnit(vtx.pvec().unit());
    const SVector<double,3> P1(vtx.track(1)->pvec());

    double P1_l = dot(P1, PtotUnit);
    // double frac     = (Pplus_l - Pminus_l) / (Pplus_l + Pminus_l);
    double pt = mag(P1 - PtotUnit * P1_l);

    float x[5];
    x[0] = phicand->mass();
    x[1] = dmin;
    x[2] = dist;
    x[3] = ctau;
    x[4] = ctau*pt;

    Nphi->Fill(x);
  } // for SDPM

  //------------ Loop over Jpsi Candidates -----------
  const TwoProngCandidateIterator JPM_end = JPsi->matches().end();
  for(TwoProngCandidateIterator PM_it = JPsi->matches().begin();
      PM_it != JPM_end; ++PM_it) {
    TwoProngCandidate jpsicand = *PM_it;

    SVertex<2>& vtx = *jpsicand->vertex();
    SVector<double,3> pos;
    vtx.findVertex3D(pos); // compute vertex manually
    vtx.set_vpos(pos);     // set vertex position manually
    SVector<double,2> rd = vtx.rmsDistAngle(); // calc rms distance
    jpsicand->calc_mother_tr();  // must be calculated after vpos is set!

    double dmin = TVdistance(*jpsicand, primvtx);     // track-vtx distance
    double dist = SGNVVdistance(*jpsicand, primvtx);  // signed vtx-vtx distance

    //    if(dist < -0.1) { jpsicand->invalid(); continue; }

    double ctau = jpsicand->mass() * dist / jpsicand->p();

    const SVector<double,3> PtotUnit(vtx.pvec().unit());
    const SVector<double,3> P1(vtx.track(1)->pvec());

    double P1_l = dot(P1, PtotUnit);
    // double frac     = (Pplus_l - Pminus_l) / (Pplus_l + Pminus_l);
    double pt = mag(P1 - PtotUnit * P1_l);

    float x[6];
    x[0] = jpsicand->mass();
    x[1] = dmin;
    x[2] = dist;
    x[3] = ctau;
    x[4] = ctau*pt;
    x[5] = rd[0];

    Njpsi->Fill(x);
  } // for SDPM

  //------------ B0: find trigger tracks -----------
  unsigned int ntri = 0, ttrig = 0;
  double pmax = 0., ptmax = 0.;
  int tpid = 0, tchrg = 0;
  if(B0->matches().size() > 0) {
    const ArteTableIterator<RecoTrack> rend = ArteTable<RecoTrack>::end();
    for(ArteTableIterator<RecoTrack> rit = ArteTable<RecoTrack>::begin();
	rit != rend; ++rit) {
      const ArtePointer<RecoTrack> rtra = *rit;
      if(rtra->cmp().slte() || rtra->cmp().sltmu()) {
	++ntri;
	if(rtra->pt() > ptmax) {
	  tchrg = rtra->charge();
	  pmax  = rtra->p();
	  ptmax = rtra->pt();
	  tpid  = rtra->pid();
	  ttrig = 0;
	  if(rtra->cmp().slte())
	    ttrig += 1;
	  if(rtra->cmp().sltmu())
	    ttrig += 2;
	}
      }
    }
  }

  //------------ Loop over B0 Candidates -----------
  const TwoProngCandidateIterator BPM_end = B0->matches().end();
  for(TwoProngCandidateIterator PM_it = B0->matches().begin();
      PM_it != BPM_end; ++PM_it) {
    TwoProngCandidate b0cand = *PM_it;

    SVertex<2>& vtx = *b0cand->vertex();
    SVector<double,3> pos;
    vtx.findVertex3D(pos); // compute vertex manually
    vtx.set_vpos(pos);     // set vertex position manually
    SVector<double,2> rd = vtx.rmsDistAngle(); // calc rms distance
    b0cand->calc_mother_tr();  // must be calculated after vpos is set!

    double dmin = TVdistance(*b0cand, primvtx);     // track-vtx distance
    double dist = SGNVVdistance(*b0cand, primvtx);  // signed vtx-vtx distance

    //    if(dist < -0.1) { b0cand->invalid(); continue; }

    double ctau = b0cand->mass() * dist / b0cand->p();

    const SVector<double,3> PtotUnit(vtx.pvec().unit());
    const SVector<double,3> P1(vtx.track(1)->pvec());

    double P1_l = dot(P1, PtotUnit);
    // double frac     = (Pplus_l - Pminus_l) / (Pplus_l + Pminus_l);
    double pt = mag(P1 - PtotUnit * P1_l);

    const RecoTrack& r1 = b0cand->PatternMatch(0)->getRecoTrack();
    const RecoTrack& r2 = b0cand->PatternMatch(1)->getRecoTrack();
    unsigned int trig0 = 0;
    unsigned int trig1 = 0;
    unsigned int trig  = 0;
    if( r1.cmp().slte() || r1.cmp().sltmu() ) ++trig0;
    if( r2.cmp().slte() || r2.cmp().sltmu() ) ++trig1;
    trig = trig0 + trig1;
    if(trig == 1) { // mark if hi-p() track is triggered
      if( (r1.p() < r2.p() && trig1 == 1) ||
	  (r2.p() < r1.p() && trig0 == 1) )
	trig = 10;
    }

    unsigned int pid = r1.pid()+r2.pid();
    if(pid == 4) { // 1 identified pion
      // flag if low mom. particle has pid
      if( (r1.p() < r2.p() && r1.pid() == 4) ||
	  (r2.p() < r1.p() && r2.pid() == 4) )
	pid = 10;
    }

    // tx, ty of hi-mom track
    double htx = 0., hty = 0.;
    if(r1.p() > r2.p()) {
      htx = r1.tx();
      hty = r1.ty();
    } else {
      htx = r2.tx();
      hty = r2.ty();
    }

    // compute B0->K+pi- mass, correct wrt charge
    double kpimass = vtx.mass_tr(SVector<double,2>(pi_minus.mass(),K_plus.mass()));
    // simply do the opposite
    double pikmass = vtx.mass_tr(SVector<double,2>(K_minus.mass(),pi_plus.mass()));

    double kpm=0.;
    if(ntri == 1) {
      if(tchrg > 0) // pos trig track from B+ -> other B is in Anti-B0 -> K-pi+
	kpm = pikmass;
      if(tchrg < 0)
	kpm = kpimass;
    }

    float x[29];
    x[0]  = b0cand->mass();
    x[1]  = dmin;
    x[2]  = dist;
    x[3]  = ctau;
    x[4]  = ctau*pt;
    x[5]  = event->run();
    x[6]  = r1.pt();
    x[7]  = r2.pt();
    x[8]  = trig;
    x[9]  = rd[0];
    x[10] = event->evt();
    x[11] = r1.p();
    x[12] = r2.p();
    x[13] = b0cand->pt();
    x[14] = b0cand->p();
    x[15] = pid;
    x[16] = rd[1];
    x[17] = fabs(b0cand->vx() - primvtx.vx());
    x[18] = fabs(b0cand->vz() - primvtx.vz());
    x[19] = pmax;
    x[20] = ptmax;
    x[21] = tpid;
    x[22] = ttrig;
    x[23] = ntri;
    x[24] = kpimass;
    x[25] = pikmass;
    x[26] = htx;
    x[27] = hty;
    x[28] = kpm;

    NB->Fill(x);
    tri->Fill(x[0], trig);

  } // for SDPM

  //------------ Loop over B0->Kpi Candidates -----------
  const TwoProngCandidateIterator BKaPM_end = B0b->matches().end();
  for(TwoProngCandidateIterator PM_it = B0b->matches().begin();
      PM_it != BKaPM_end; ++PM_it) {
    TwoProngCandidate b0cand = *PM_it;

    SVertex<2>& vtx = *b0cand->vertex();
    SVector<double,3> pos;
    vtx.findVertex3D(pos); // compute vertex manually
    vtx.set_vpos(pos);     // set vertex position manually
    SVector<double,2> rd = vtx.rmsDistAngle(); // calc rms distance
    b0cand->calc_mother_tr();  // must be calculated after vpos is set!

    double dmin = TVdistance(*b0cand, primvtx);     // track-vtx distance
    double dist = SGNVVdistance(*b0cand, primvtx);  // signed vtx-vtx distance

    //    if(dist < -0.1) { b0cand->invalid(); continue; }

    double ctau = b0cand->mass() * dist / b0cand->p();

    const SVector<double,3> PtotUnit(vtx.pvec().unit());
    const SVector<double,3> P1(vtx.track(1)->pvec());

    double P1_l = dot(P1, PtotUnit);
    // double frac     = (Pplus_l - Pminus_l) / (Pplus_l + Pminus_l);
    double pt = mag(P1 - PtotUnit * P1_l);

    const RecoTrack& r1 = b0cand->PatternMatch(0)->getRecoTrack();
    const RecoTrack& r2 = b0cand->PatternMatch(1)->getRecoTrack();

    unsigned int trig0 = 0;
    unsigned int trig1 = 0;
    unsigned int trig  = 0;
    if( r1.cmp().slte() || r1.cmp().sltmu() ) ++trig0;
    if( r2.cmp().slte() || r2.cmp().sltmu() ) ++trig1;
    trig = trig0 + trig1;
    if(trig == 1) { // mark if hi-p() track is triggered
      if( (r1.p() < r2.p() && trig1 == 1) ||
	  (r2.p() < r1.p() && trig0 == 1) )
	trig = 10;
    }

    float x[15];
    x[0]  = b0cand->mass();
    x[1]  = dmin;
    x[2]  = dist;
    x[3]  = ctau;
    x[4]  = ctau*pt;
    x[5]  = event->run();
    x[6]  = r1.pt();
    x[7]  = r2.pt();
    x[8]  = trig;
    x[9]  = rd[0];
    x[10] = r1.p();
    x[11] = r2.p();
    x[12] = b0cand->pt();
    x[13] = b0cand->p();
    x[14] = r1.pid() + r2.pid();

    NBkp->Fill(x);
  }


  //------------ Loop over B0bar->Kpi Candidates -----------
  const TwoProngCandidateIterator BKbPM_end = B0bar->matches().end();
  for(TwoProngCandidateIterator PM_it = B0bar->matches().begin();
      PM_it != BKbPM_end; ++PM_it) {
    TwoProngCandidate b0barcand = *PM_it;

    SVertex<2>& vtx = *b0barcand->vertex();
    SVector<double,3> pos;
    vtx.findVertex3D(pos); // compute vertex manually
    vtx.set_vpos(pos);     // set vertex position manually
    SVector<double,2> rd = vtx.rmsDistAngle(); // calc rms distance
    b0barcand->calc_mother_tr();  // must be calculated after vpos is set!

    double dmin = TVdistance(*b0barcand, primvtx);     // track-vtx distance
    double dist = SGNVVdistance(*b0barcand, primvtx);  // signed vtx-vtx distance

    //    if(dist < -0.1) { b0barcand->invalid(); continue; }

    double ctau = b0barcand->mass() * dist / b0barcand->p();

    const SVector<double,3> PtotUnit(vtx.pvec().unit());
    const SVector<double,3> P1(vtx.track(1)->pvec());

    double P1_l = dot(P1, PtotUnit);
    // double frac     = (Pplus_l - Pminus_l) / (Pplus_l + Pminus_l);
    double pt = mag(P1 - PtotUnit * P1_l);

    const RecoTrack& r1 = b0barcand->PatternMatch(0)->getRecoTrack();
    const RecoTrack& r2 = b0barcand->PatternMatch(1)->getRecoTrack();

    unsigned int trig0 = 0;
    unsigned int trig1 = 0;
    unsigned int trig  = 0;
    if( r1.cmp().slte() || r1.cmp().sltmu() ) ++trig0;
    if( r2.cmp().slte() || r2.cmp().sltmu() ) ++trig1;
    trig = trig0 + trig1;
    if(trig == 1) { // mark if hi-p() track is triggered
      if( (r1.p() < r2.p() && trig1 == 1) ||
	  (r2.p() < r1.p() && trig0 == 1) )
	trig = 10;
    }

    float x[15];
    x[0]  = b0barcand->mass();
    x[1]  = dmin;
    x[2]  = dist;
    x[3]  = ctau;
    x[4]  = ctau*pt;
    x[5]  = event->run();
    x[6]  = r1.pt();
    x[7]  = r2.pt();
    x[8]  = trig;
    x[9]  = rd[0];
    x[10] = r1.p();
    x[11] = r2.p();
    x[12] = b0barcand->pt();
    x[13] = b0barcand->p();
    x[14] = r1.pid() + r2.pid();

    NBkp->Fill(x);
  }

  return true;
}

//==============================================================================
// EventCleanup
//==============================================================================
bool v0Analysis::EventCleanup() {
  return true;
}

//==============================================================================
// End processing
//==============================================================================
bool v0Analysis::End() {

  cout << "Accepted events: " << nacc
       << " rejected: " << nrej
       << " fraction: " << nacc/static_cast<double>(nacc+nrej)
       << endl
       << "Tracks with PID: " << pidyes
       << " without PID: " << pidno
       << " fraction: " << pidyes/static_cast<double>(pidyes+pidno)
       << endl;
  return true;
}

