#ifndef _OBJ_ARRAY_HH_
#define _OBJ_ARRAY_HH_
// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   03. Sep 1997
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Template wrapper for TObjArray class
//
// changes:
// 03 Sep 1997 creation by C. Borgmeier
// 13 Nov 2000 (TG) added doc++ comments
// 14 Nov 2000 (TG) declared some member functions inline
// 20 Nov 2000 (TG) added namespace CLUE
// 05 Dez 2000 (TG) added operator<< for ConstGObjArrayIterator
// 05 Jan 2001 (TG) added operator<() in [Const]GObjArrayIterator
// 14 Nov 2001 (TG) code polishing
//
// ********************************************************************

#ifdef USE_ROOT

#include <iterator>
#include <iosfwd>
#include <TObjArray.h>


#ifdef NAMESPACE
#include "gate/RtypesT.h"

namespace CLUE {
#endif

  template <class T> class GObjArray;
  template <class T> class GObjArrayIterator;
  template <class T> class ConstGObjArrayIterator;
  template <class T> class GClonesArray;


  /** GObjArray is a STL compatible template wrapper for the ROOT TObjArray class.
      This class is only available in ROOT. Outside ROOT, GObjarray is typedef'd
      to a STL vector.

      \textbf{Note:} GObjArray assumes that the container type is a pointer type!

      @memo Array of different objects.
      @author C. Borgmeier
  */
  //==============================================================================
  // GObjArray
  //==============================================================================
  template<class T>
    class GObjArray: public TObjArray {
    public:
      /** @name --- Typedefs --- */
      ///
      typedef T                         value_type;
      ///
      typedef value_type&               reference;
      ///
      typedef const value_type&         const_reference;
      ///
      typedef size_t                    size_type;
      ///
      typedef GObjArrayIterator<T>      iterator;
      ///
      typedef ConstGObjArrayIterator<T> const_iterator;

      /** @name --- Constructors --- */
      ///
      GObjArray();
      ///
      GObjArray( const GObjArray<T>& rhs );
      ///
      ~GObjArray();
      ///
      GObjArray<T>& operator=( const GObjArray<T>& rhs );

      /** @name --- Access methods --- */
      /// begin iterator
      iterator       begin();
      ///
      const_iterator begin() const;
      /// end iterator
      iterator       end();
      ///
      const_iterator end() const;

      /// size of container
      size_type size() const;
      /// is container empty?
      bool empty() const;
      
      /// add object to container
      void push_back( value_type obj );
      /// remove last element
      void pop_back();
      /// erase element at iterator position
      void erase( iterator position );
      /// erase elements in iterator range
      void erase( iterator p1, iterator p2 );
      /// clear container
      void clear();
      
      /// readonly element access via index
      const value_type operator[]( int idx ) const;
      /// element access via index
      value_type GetAt( int idx );
      /// readonly element access via index
      const value_type GetAt( int idx ) const;

      /// set a value via index
      void set( int idx, value_type v );

      /** @name --- Expert methods --- */
      /// used by operator<<()
      std::ostream& print( std::ostream& os ) const;
      ///
      void Print( Option_t* ="" );
      ///  put elements of this container into newArray container
      void saveObjects( GObjArray<T>& newArray ) const;
  
      ClassDefT(GObjArray<T>,1)
  }; // end of class GObjArray


  //==============================================================================
  // operator <<
  //==============================================================================
  template<class T>
    inline std::ostream& operator<<( std::ostream& os, const GObjArray<T>& ar ) {
    return ar.print( os );
  }

  
  /** Non-const iterator class for GObjectArray and GClonesArray.

      @memo GObjectArray Iterator class.
      @author C. Borgmeier, T. Glebe
  */
  //==============================================================================
  // GObjArrayIterator
  //==============================================================================
  template<class T>
    class GObjArrayIterator: public TObject {
    public:
      /** @name --- Typedefs --- */
      ///
      typedef T                    value_type;
      ///
      typedef T*                   pointer;
      ///
      typedef T&                   reference;
      ///
      typedef const T&             const_reference;
      ///
      typedef forward_iterator_tag iterator_category;
      ///
      typedef ptrdiff_t            difference_type;

      /** @name --- Constructors --- */
      ///
      GObjArrayIterator();
      ///
      GObjArrayIterator( TObjArray& array, Int_t idx );
      ///
      GObjArrayIterator( const GObjArrayIterator<T>& rhs );
      ///
      const GObjArrayIterator<T>& operator=( const GObjArrayIterator<T>& rhs );

      /** @name --- Operators --- */
      ///
      value_type operator*();
      ///
      const value_type operator*() const;
      ///
      GObjArrayIterator<T>& operator++();
      ///
      GObjArrayIterator<T> operator++(int);
      ///
      GObjArrayIterator<T>& operator--();
      ///
      GObjArrayIterator<T> operator--(int);
      ///
      GObjArrayIterator<T> operator+(int i) const;

      ///
      bool operator!=( const ConstGObjArrayIterator<T>& rhs ) const;
      ///
      bool operator==( const GObjArrayIterator<T>& rhs ) const;
      ///
      bool operator<( const GObjArrayIterator<T>& rhs ) const;

      /** @name --- Expert functions --- */
      ///
      void Print( Option_t* o="" );
      /// used by operator<<()
      std::ostream& print( std::ostream& os ) const;

    protected:
      ///
      void increment();
      ///
      void decrement();
      
    public:
      TObjArray* fArray;
      Int_t fCursor;
      
      ClassDefT(GObjArrayIterator<T>,1)
  }; // end of class GObjArrayIterator


  //==============================================================================
  // operator <<
  //==============================================================================
  template<class T>
    inline std::ostream& operator<< (std::ostream& os,
				     const GObjArrayIterator<T>& it ) {
    return it.print(os);
  }
  

  /** Const iterator class for GObjectArray and GClonesArray.

      @memo ConstGObjectArray Iterator class.
      @author C. Borgmeier, T. Glebe
  */
  //==============================================================================
  // ConstGObjArrayIterator
  //==============================================================================
  template<class T>
    class ConstGObjArrayIterator : public TObject {
      friend class GObjArrayIterator<T>;
    public:
      /** @name --- Typedefs --- */
      ///
      typedef T                         value_type;
      ///
      typedef value_type*               pointer;
      ///
      typedef value_type&               reference;
      ///
      typedef const value_type&         const_reference;
      ///
      typedef forward_iterator_tag      iterator_category;
      ///
      typedef ptrdiff_t                 difference_type;

      /** @name --- Constructors --- */
      ///
      ConstGObjArrayIterator();
      ///
      ConstGObjArrayIterator( const ConstGObjArrayIterator<T>& rhs );
      ///
      ConstGObjArrayIterator( const GObjArrayIterator<T>& rhs );
      /// this would better be protected.
      ConstGObjArrayIterator( const TObjArray& array, Int_t idx );

      /** @name --- Operators --- */
      ///
      const ConstGObjArrayIterator<T>& operator=( const
						  ConstGObjArrayIterator<T>&
						  rhs );
      ///
      const ConstGObjArrayIterator<T>& operator=( const GObjArrayIterator<T>& rhs );
      ///
      value_type operator*();
      ///
      const value_type operator*() const;
      ///
      ConstGObjArrayIterator<T>& operator++();
      ///
      ConstGObjArrayIterator<T> operator++(int);
      ///
      ConstGObjArrayIterator<T>& operator--();
      ///
      ConstGObjArrayIterator<T> operator--(int);
      ///
      ConstGObjArrayIterator<T> operator+(int i) const;
      ///
      bool operator!=( const ConstGObjArrayIterator<T>& rhs ) const;
      ///
      bool operator==( const ConstGObjArrayIterator<T>& rhs ) const;
      ///
      bool operator<( const GObjArrayIterator<T>& rhs ) const;

      /** @name --- Expert methods --- */
      /// used by operator<<()
      std::ostream& print( std::ostream& os ) const;

    protected:
      ///
      void increment();
      ///
      void decrement();
      
      TObjArray* fArray;
      Int_t fCursor;

      ClassDefT(ConstGObjArrayIterator<T>,1)
  }; // end of class ConstGObjArrayIterator


  //==============================================================================
  // operator <<
  //==============================================================================
  template<class T>
    inline std::ostream& operator<< (std::ostream& os,
				     const ConstGObjArrayIterator<T>& it ) {
    return it.print(os);
  }
  

#ifdef NAMESPACE
} // namespace CLUE

MyClassDefT2(CLUE,GObjArray,T)
  ; // for doc++
MyClassDefT2(CLUE,GObjArrayIterator,T)
  ; // for doc++
MyClassDefT2(CLUE,ConstGObjArrayIterator,T)
  ; // for doc++
#else
ClassDefT2(GObjArray,T)
  ; // for doc++
ClassDefT2(GObjArrayIterator,T)
  ; // for doc++
ClassDefT2(ConstGObjArrayIterator,T)
  ; // for doc++
#endif


#else  // USE_ROOT
#include <vector>

#define GObjArray              std::vector
#endif // USE_ROOT

#ifndef __CINT__
#include "gate/ObjArray.icc"
#endif
#endif
