#ifndef __GSCOMBINER_HH
#define __GSCOMBINER_HH
// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   05. Jun 2001
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: template for a container combiner. GSCombiner is a combination
//              of SDCombiner with GCombiner.
//
// changes:
// 05 Jun 2001 (TG) creation from GCombiner.
// 02 Jul 2001 (TG) push_back(void*) added
// 03 Jul 2001 (TG) push_back(void*) removed
// 05 Jul 2001 (TG) GSCombinerIterator::iterator, ::const_iterator typedefs added
// 09 Jul 2001 (TG) non-const begin(), end() added
//
// *****************************************************************************

#include <iosfwd>
#include <list>
#include <vector>
#include "SDCombiner.hh"

template <class T> class GSCombiner;
template <class T> class GSCombinerIterator;
template <class T> class GSUniqueCombinerIterator;

/** GSCombiner helper class.
    This class performs the bookkeeping of the begin, end and current iterator
    for each container saved by the Combiner class. The running interator
    can be resetted to the begin of the container in case it reached the end of
    the container. This class avoids that one has to declare the
    \texttt{GSCombinerIterator} as a friend of the \texttt{GSCombiner} class and
    thus reduces the coupling between those classes.

    \textbf{Note:} This class is not for public use.
    @memo GSCombiner helper class
    @author T. Glebe
 */
//============================================================================
// Class GIteratorPair - class for iterator range
//============================================================================
template <class Iterator>
class GIteratorPair {
  ///
  typedef GIteratorPair<Iterator> self;

public:
  /** @name --- Constructors --- */
  ///
  GIteratorPair() {}
  ///
  GIteratorPair(const GIteratorPair& rhs) : 
    begin(rhs.begin), end(rhs.end), iter(rhs.iter) {}
  ///
  GIteratorPair(const Iterator& a) : 
    begin(a), end(a), iter(a) {}
  ///
  GIteratorPair(const Iterator& lower, const Iterator& upper) :
    begin(lower), end(upper), iter(lower) {}

  /** @name --- Access methods --- */
  /// has iter reached end?
  inline bool iteratable() const { return iter != end; }
  /// do an iteration step
  inline self& iterate() { ++iter; return *this; }

  /** @name --- Operators --- */
  ///
  inline bool operator==(const GIteratorPair<Iterator>& rhs) const {
    return iter == rhs.iter;
  }
  ///
  inline bool operator!=(const GIteratorPair<Iterator>& rhs) const {
    return iter != rhs.iter;
  }
  ///
  inline bool operator<(const GIteratorPair<Iterator>& rhs) const {
    return iter < rhs.iter;
  }

  /** @name --- Expert functions --- */
  ///
  inline void reset() { iter = begin; }
  ///
  inline std::ostream& print(std::ostream& os) const { return os << iter; }
  ///
  inline const Iterator& getIter() const { return iter; }

private:
  Iterator begin;  // lower bound
  Iterator end;    // upper bound
  Iterator iter;
}; // class GIteratorPair

//==============================================================================
// operator<<
//==============================================================================
template<class Iterator>
inline std::ostream& operator<<(std::ostream& s, const GIteratorPair<Iterator>& a) {
  return a.print(s);
}

/** Container class.
    The Combiner class stores pointers to containers whose elements should
    be combined. The number of containers is unlimited.
    @memo Combiner
    @author T. Glebe
*/
//============================================================================
// Class GSCombiner - container for container references
//============================================================================
template <class T>
class GSCombiner {
  typedef typename GC_trait<T>::Gvalue_type             Gvalue_type;
  typedef typename GC_trait<T>::Gpointer                Gpointer;
  //  typedef typename GC_trait<T>::GCombiner_v             GCombiner_v;
  typedef std::list<SDCombiner<T> >                     GCombiner_v;
  typedef GCombiner_v::iterator                         GCombiner_it;
  typedef GCombiner_v::const_iterator                   GCombiner_cit;
  typedef GCombiner_v::const_iterator                   citerator;
  //  typedef typename GC_trait<T>::GCombiner_v::value_type value_type;
  //    typedef typename GC_trait<T>::GCombiner_cit           GCombiner_cit;
  //    typedef typename GC_trait<T>::GCombiner_cit           citerator;

public:
  /** @name --- Typedef's --- */
  ///
  typedef GSCombinerIterator<T>             iterator;
  ///
  typedef GSCombinerIterator<T>             const_iterator;

  /** @name --- Constructors --- */
  ///
  GSCombiner();
  ///
  GSCombiner(const GSCombiner<T>& rhs);
  ///
  GSCombiner(const Gpointer[], const unsigned int N);
  ///
  ~GSCombiner();

  /** @name --- Access functions --- */
  ///
  inline void push_back(const Gvalue_type& rel, unsigned int N) {
    rellist.push_back(SDCombiner<T>(const_cast<Gvalue_type&>(rel),N));
  }
  ///
  void push_back(const Gvalue_type& rel);
  /// no of containers used for combinatorics
  inline int  size() const { return rellist.size(); }
  ///
  inline bool empty() const { return rellist.empty(); }
  /// check if combiner is valid: true if each container size is > 0
  bool valid() const;

  /** @name --- Iterator methods --- */
#ifndef __CINT__
  /// to iterate container container, not available in CINT
  inline    citerator cbegin() const { return rellist.begin(); }
  /// to iterate container container, not available in CINT
  inline    citerator cend()   const { return rellist.end(); }
#endif
  /// to iterate combinations
  inline     iterator begin()  const { return iterator(rellist, false); }
  /// to iterate combinations
  inline     iterator end()    const { return iterator(rellist, true); }

  /// direct access
  Gpointer operator[](unsigned int i) const;

  /** @name --- Operators --- */
  ///
  GSCombiner<T>& operator=(GSCombiner<T>& rhs);
  ///
  inline bool operator==(const GSCombiner<T>& rhs) const {
    return rellist == rhs.rellist;
  }

  /** @name --- Expert functions --- */
  ///
  std::ostream& print(std::ostream& os) const;
  /// empty combiner
  inline void clear(void) { rellist.clear(); }
  /// compute number of combinations
  unsigned int ncomb(void) const;
  
protected:
  GCombiner_v  rellist;
}; // class GSCombiner


//==============================================================================
// operator<<
//==============================================================================
template<class T>
inline std::ostream& operator<<(std::ostream& s, const GSCombiner<T>& a) {
  return a.print(s);
}


/** Combination iterator.
    The \texttt{GCombinerIterator} iterates over all combinations which are
    possible by the given container objects. The \texttt{GCombinerIterator}
    represents a vector of iterators of the container objects. Therefore this
    iterator behaves very different from a STL iterator and is not
    STL conform! The \texttt{GCombinerIterator} itself has a \texttt{begin()}
    and \texttt{end()} operator to loop over the elements of the current
    combination. An \texttt{operator*()} for receiving the vector of container
    iterators and an \texttt{operator[]()} method for direct access to the
    container iterators is provided.
    @memo Iterates the combinations.
    @author T. Glebe
 */ 
//============================================================================
// Class GSCombinerIterator: non conform to STL!!!
//============================================================================
template <class T>
class GSCombinerIterator {
  
  typedef GSCombinerIterator<T> self;
#ifndef __CINT__
  typedef std::list<SDCombiner<T> >                    GCombiner_v;
  typedef GCombiner_v::const_iterator                  GCombiner_cit;
  typedef SDCombinerIterator<T>                        GIterator;
  typedef GIteratorPair<SDCombinerIterator<T> >         GItPair;
  //  typedef GIteratorPair<T>                             GItPair;
  //  typedef typename GC_trait<T>::GIterator_v        GIterator_v;
  typedef std::list<GItPair>                           GIterator_v;
  typedef GIterator_v::iterator                        GIterator_it;
  typedef GIterator_v::const_iterator                  GIterator_cit;
  //    typedef typename GC_trait<T>::GIterator_it       GIterator_it;
  //    typedef typename GC_trait<T>::GIterator_cit      GIterator_cit;
  //  typedef typename GC_trait<T>::ItPair             ItPair;

  typedef typename GC_trait<T>::GCombination       GCombination;
  typedef typename GC_trait<T>::GCombination_it    GCombination_it;
  typedef typename GC_trait<T>::GCombination_cit   GCombination_cit;
#else
  typedef GC_trait<T>::GCombiner_v        GCombiner_v;
  typedef GC_trait<T>::GCombiner_cit      GCombiner_cit;
  typedef GC_trait<T>::GIterator          GIterator;
  typedef GC_trait<T>::GIterator_v        GIterator_v;
  typedef GC_trait<T>::GIterator_it       GIterator_it;
  typedef GC_trait<T>::GIterator_cit      GIterator_cit;
  typedef GC_trait<T>::ItPair             ItPair;

  typedef GC_trait<T>::GCombination       GCombination;
  typedef GC_trait<T>::GCombination_it    GCombination_it;
  typedef GC_trait<T>::GCombination_cit   GCombination_cit;
#endif

public:
  ///
  typedef GCombination_it  iterator;
  ///
  typedef GCombination_cit const_iterator;

  /** @name --- Constructors --- */
  ///
  GSCombinerIterator() : getcur(false), size_(0) {}
  ///
  GSCombinerIterator(const GSCombinerIterator<T>& rhs) :
    iter(rhs.iter), curr(rhs.curr), last(--iter.end()), getcur(false), size_(rhs.size_) {}
  ///
  GSCombinerIterator(const GCombiner_v& v, bool isend = false) :
    getcur(false), size_(0) {
    const GCombiner_cit cend  = v.end();
    const GCombiner_cit clast = --v.end();
    for(GCombiner_cit it = v.begin(); it != cend; ++it) {
      size_ += it->N();
      if(isend)
	if(it == clast)
	  iter.push_back(GItPair(it->end()));
	else
	  iter.push_back(GItPair(it->begin()));
      else
	iter.push_back(GItPair(it->begin(), it->end()));
    } // for GCombiner
    last = --iter.end(); // remember last element
  }
  ///
  virtual ~GSCombinerIterator() {}

  /** @name --- Access methods --- */
  ///
  inline GCombination_it begin() { getCombination(); return curr.begin(); }
  ///
  inline GCombination_it end()   { getCombination(); return curr.end(); }
  ///
  inline GCombination_cit begin() const { getCombination(); return curr.begin(); }
  ///
  inline GCombination_cit end()   const { getCombination(); return curr.end(); }
  /// no of elements in a combination
  inline unsigned int size() const { return size_; }

  /** @name --- Operators --- */
  ///
  inline self&  operator++() { iterate(); return *this; }
  /// return vector with iterators of current combination
  inline const GCombination&  operator*() const { getCombination(); return curr; }
  /// access iterator $i$ of current combination
  inline const typename GC_trait<T>::GIterator& operator[](unsigned int i) const { 
    getCombination(); return curr[i];
  }
  ///
  inline bool operator==(const GSCombinerIterator<T>& rhs) const {
    return iter == rhs.iter;
  }
  ///
  inline bool operator!=(const GSCombinerIterator<T>& rhs) const {
    return iter != rhs.iter;
  }

  /** @name --- Expert functions --- */
  /// used by operator<<()
  std::ostream& print(std::ostream& os) const;

protected:
  /// algorithm for an iteration step
  virtual void iterate();
  ///
  void getCombination() const;

  GIterator_v           iter;   // vector of container iterators
  mutable GCombination  curr;   // vector of current iterators
  GIterator_it          last;   // last iterator in iter
  mutable bool          getcur; // getCombination() status variable
  unsigned int          size_;  // no of elements in combination
}; // class GSCombinerIterator


//==============================================================================
// operator<<
//==============================================================================
template<class T>
inline std::ostream& operator<<(std::ostream& s, const GSCombinerIterator<T>& a) {
  return a.print(s);
}


//============================================================================
// load the implementation file
//============================================================================
#ifndef __CINT__
#include "GSCombiner.icc"
#endif
#endif
