#include <algorithm>

#ifndef GROVER_WITHOUT_ARTE
#include "arte/arte.hh"
#include "arte/RSEG_cmp.hh"
#endif

#include "grover/grover.hh"
#include <map.h>

//
// Part of the Grover package
// Please see the header file grover.hh
// for authors etc.
//
//
// Modifications:
// When?     Who?     What?
// YY-MM-DD  Name     Modification
// -----------------------------------------------------------------------------
// 01-03-13  Max      Added preprocessor directives
//

int GroverFillLikeMD(GroverParticleList* gpl)
{
  /////////////////////////////////////////////////////////////////////////
  //
  // Fills a list using the method in Michael Dressel's thesis analysis:
  //
  // Fills a GroverParticleList using all VDS tracks with hsi >= MINHITS_VDS 
  // giving priority to VDS OTR tracks according to Michael Schmellings
  // track selection for GMB data
  // use all RTRAs with hsi >= MINHITS_VDS && hou >= MINHITS_OTR also use
  // all RTRAs  with hsi VDS >= MINHITS_VDS 
  // give weights: tracks with VDS and OTR get the highest weight
  //               weighting among VDS OTR tracks according
  //               to Michael Schmelling
  // weight all RTRAs that have a VDS segment 
  //               similar to Michael Schmelling 
  // do clone removal such that all VDS segments are kept.
  //
  // It happens, that if two VDS tracks are matched with the same OTR track
  // and for one of that tracks this match is the only RTRA entry,
  // both tracks are used. So one OTR track could be used twice.
  //
  // Author: Michael Dressel
  // Date:   22 Nov 2000
  //
  /////////////////////////////////////////////////////////////////////////

  int errorflag=SUCCESS;

#ifndef GROVER_WITHOUT_ARTE
  const int VDSOTR_OFFSET = 10000;
  const int MINHITS_VDS   = 6;
  const int MINHITS_OTR   = 20;

  int ntr,i;

  vector< ArtePointer<RTRA> > rtra;

  map< float, ArtePointer<RTRA> > mrtra;

  //------------------------------------
  // put all VDS-OTR with offset VDSOTR_OFFSET  and all
  // VDS tracks into a sorted map

  for(ArteTable<RTRA>::iterator itrtra = ArteTable<RTRA>::begin(); 
      itrtra != ArteTable<RTRA>::end(); ++itrtra) {
    if( (*itrtra)->fit != 0 ) continue;
    int nvds  = (*itrtra)->hsi;
    int notr  = (*itrtra)->hou;
    
    if (nvds >= MINHITS_VDS && notr >= MINHITS_OTR ) {
      // VDS-OTR tracks
      float fom = float(nvds+notr) + 1./(1.+(*itrtra)->chi) + VDSOTR_OFFSET;
      mrtra.insert(pair< float, ArtePointer<RTRA> >(-fom,*itrtra));
    } 
    if (nvds >= MINHITS_VDS) {
      // VDS tracks
      float fom = float(nvds) + 1./(1.+(*itrtra)->chi);
      mrtra.insert(pair< float, ArtePointer<RTRA> >(-fom,*itrtra));
    }
  } 


  // copy the tracks into the return vector, omitting all clones 

  for(map< float, ArtePointer<RTRA> >::iterator it=mrtra.begin(); 
      it != mrtra.end(); ++it) {
    int OK = 1;
    for(unsigned int i=0; i!=rtra.size(); ++i) {
      if ((-1* (it->first)) >VDSOTR_OFFSET) {
	if(rtra[i]->csvd  == it->second->csvd ) OK=0;
	if(rtra[i]->cmain == it->second->cmain) OK=0;
      } else {
	if(rtra[i]->csvd  == it->second->csvd ) OK=0;
      }	
      if(!OK) break;
    }
    if(OK) {
      rtra.push_back(it->second);
    }
  }

  ntr = rtra.size();

  for(i=0; i!=ntr; i++) {
    GroverParticle temp(rtra[i]);
    gpl->push_back(temp);
  }
  if (gpl->empty()) errorflag = FAILURE;

  /* check if all VDS tracks with hsi>=6 are used. ok
  vector< ArtePointer<RTRA> > rtravds;

  for(map< float, ArtePointer<RTRA> >::iterator it=mrtra.begin(); 
      it != mrtra.end(); ++it) {
    int OK = 1;
    for(unsigned int i=0; i!=rtravds.size(); ++i) {
      if(rtravds[i]->csvd  == it->second->csvd ) OK=0;
      if(!OK) break;
    }
    if(OK) {
      rtravds.push_back(it->second);
    }
  }
  if (rtravds.size() != ntr) {
    cout << "GroverFillLikeMD: ERROR " << endl;
    cout << rtravds.size() << " " << ntr << endl;
  }
  */
#endif


  return errorflag;
}

