#ifndef __PATTERNIF_HH
#define __PATTERNIF_HH
// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   28. Feb 2001
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Pattern object interface
//
// changes:
// 22 Aug 2001 (TG) creation
// 23 Aug 2001 (TG) added setProperty(), getDecayString(), printProperty()
// 24 Aug 2001 (TG) added fillMC()
// 27 Aug 2001 (TG) added isReconstructed(MC_Track), fillRECOfromMC()
// 28 Aug 2001 (TG) added isNotReconstructed(MC_Track), fillNORECOfromMC()
//
// ********************************************************************
#include <iosfwd>
#include "pattern/P_trait.hh"

class TrackIf;
class MC_Track;
class ToppIf;
class PatternMatchIf;
template <class C, class T> class SProperty;

/** Abstract class which defines the minimal interface for a Pattern object.

    @memo Pattern Interface
    @author T. Glebe
*/
//==============================================================================
// PatternIf
//==============================================================================
class PatternIf {
public:
  /** @name --- Access methods --- */
  /// check if single Track passes property criterias
  virtual bool evalTrack(const TrackIf& rhs) const = 0;
  /// check track property without charge
  virtual bool evalTrackProp(const TrackIf& rhs) const = 0;
  /// set property object
  virtual void setProperty(const SProperty<TrackIf,bool>& rhs) = 0;
  /// get topp object
  virtual const ToppIf& topp() const = 0;
  /// get array to rest-masses
  virtual const double* rm() const = 0;
  /// get length of rest-mass array
  virtual unsigned int len() const = 0;
  /// return string with decay
  virtual string getDecayString(unsigned int lvl=0) const = 0;
  /// print property string
  virtual void printProperty(std::ostream& os, unsigned int lvl=0) const = 0;
  /// get PatternMatchContainer pointers
  virtual const SPMP_l* getContainer() const = 0;
  /// fill container from MC event
  virtual const PatternMatchIf* fillMC(const MC_Track& mt) const = 0;
  /// fill container with reconstructed tracks from MC event
  virtual PatternMatchIf* fillRECOfromMC(const MC_Track& mt) const = 0;
  /// fill container with reconstructed tracks or (if not reconstr.) MC tracks from MC event
  virtual PatternMatchIf* fillNORECOfromMC(const MC_Track& mt) const = 0;

  /** @name --- Operators --- */
  /// compare with MC track
  virtual bool operator==(const MC_Track& rhs) const = 0;

  /** @name --- Expert methods --- */
  /// same as operator==(MC\_Track), but also check whether MC\_Track was reconstructed
  virtual bool isReconstructed(const MC_Track& rhs) const = 0;
  /// same as operator==(MC\_Track), but also check if MC\_Track was not reconstructed
  virtual bool isNotReconstructed(const MC_Track& rhs, bool& no_break) const = 0;
  /// used by operator<<()
  virtual std::ostream& print( std::ostream& ) const = 0;
}; // end of class PatternMatchIf


//==============================================================================
// operator <<
//==============================================================================
inline std::ostream& operator<< (std::ostream& os, const PatternIf& t) {
  return t.print(os);
}
#endif
