#ifndef __SPATTERNMATCHCONTAINER_HH
#define __SPATTERNMATCHCONTAINER_HH
// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   30. Apr 2001
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: A PatternMatch container class
//
// changes:
// 30 Apr 2001 (TG) creation
// 12 Jun 2001 (TG) small bug fix in operator<<(SmartPointer...)
// 28 Jun 2001 (TG) removed template parameter C and T
// 29 Jun 2001 (TG) added push_back(RecoTrack)
// 02 Jul 2001 (TG) added ptrlist, added pmatches()
// 22 Aug 2001 (TG) added with(), removed const from pattern_, changed return
//                  type of pattern() to SmartPointer
// 23 Aug 2001 (TG) changed push_back() return type
// 28 Aug 2001 (TG) changed push_back(TrackIf) to push_back(MC_Track)
// 05 Nov 2001 (TG) added push_back(TrackIf)
// 14 Jan 2002 (TG) added ownership argument to push_back(RecoSegment)
// 15 Jan 2002 (TG) added exist() function from SPattern.hh
//
// ********************************************************************

#include <iosfwd>
#include "pattern/SmartPointer.hh"
#include "pattern/P_trait.hh"

class SPattern;
class RecoSegment;
class MC_Track;
class TrackIf;
class SPatternMatch;

/** Class which represents a container for Pattern matches.

    @memo Generic class
    @author T. Glebe
*/
//==============================================================================
// SPatternMatchContainer
//==============================================================================
class SPatternMatchContainer {

public:
  /** @name --- Constructors --- */
  ///
  SPatternMatchContainer();
  ///
  SPatternMatchContainer(const SPattern& rhs);
  ///
  SPatternMatchContainer(const SmartPointer<SPattern>& rhs);
  ///
  ~SPatternMatchContainer();

  /** @name --- Typedefs --- */
  ///
  typedef SPM_it  iterator;
  ///
  typedef SPM_cit const_iterator;

  /** @name --- Access methods --- */
  /** Add a RecoSegment. If ownership == \texttt{true}, the SPatternMatch
      destructor will delete the RecoSegment object. This is useful in case the
      RecoSegment object was created via the
      \texttt{SDecayPatternMatch::createMotherSegment()} method.*/
  ParticleCandidate push_back( RecoSegment& rs, bool ownership = false);
  /// add a MC track
  ParticleCandidate push_back( MC_Track& mtra);
  /// add a TrackIf
  ParticleCandidate push_back( TrackIf& t );
  /// clear list of pattern matches
  void clear();
  /// no. of entries in container
  unsigned int size() const;
  /// readonly access to pattern matches
  const SPM_l& matches() const;
  /// read/write access to pattern matches
  SPM_l& matches();
  /// readonly access to pattern match pointers
  const SPMP_l& pmatches() const;
  /// read/write access to pattern match pointers
  SPMP_l& pmatches();
  /// get pattern
  const SP_sp pattern() const;
  /// set a Track Property
  void with(const SProperty<TrackIf,bool>& rhs);

  /** @name --- Expert methods --- */
  /// check track property without charge
  bool evalTrackProp(const TrackIf& rhs) const;
  /** Find pattern matches from Input container. \textbf{Note:} it is assumed
      that the input container stores pointers to the track objects! */
  template <class Input>
  bool select(const Input& inp);
  /** Return \texttt{true} if a track which matches the pattern exists in
      InputContainer. \textbf{Note:} is is assumed that InputContainer stores
      pointers to the track objects! */
  template <class InputContainer>
  bool exist(const InputContainer& src) const;
  /// used by operator<<()
  std::ostream& print(std::ostream& os) const;

private:
  SPM_l       pmlist;
  SPMP_l      ptrlist;
  SP_sp       pattern_;
};

//==============================================================================
// operator <<: the second template argument is a problem, therefore hardcoded
//==============================================================================
inline std::ostream& operator<<( std::ostream& os, const SPatternMatchContainer& a ) {
  return a.print( os );
}

//==============================================================================
// operator <<: the second template argument is a problem, therefore hardcoded
//==============================================================================
inline std::ostream& operator<<( std::ostream& os, const
				 SmartPointer<SPatternMatchContainer>& a ) {
  return a.null() ? os << "(nil)" : a->print( os );
}

#include "pattern/SPatternMatchContainer.icc"
#endif
