#ifndef __GDECAYPATTERNBASE_HH
#define __GDECAYPATTERNBASE_HH
// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   31. Jan 2001
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Class describing a decay pattern, base part.
//              This base part is only responsible for saving the
//              track patterns.
//
// changes:
// 31 Jan 2001 (TG) creation
// 07 Feb 2001 (TG) changed topp to topp_, added topp() member function
// 23 Feb 2001 (TG) bug fix in topp()
// 05 Mar 2001 (TG) getDecayString(), getBasicDecayString(), 
//                  getPatternString(), getBasicPatternString() added
// 05 Mar 2001 (TG) flip_to_cc(), eval_patlist(), operator==(ToppIf) added
// 08 Mar 2001 (TG) added dm and massrange() from GDecayPattern
// 09 Mar 2001 (TG) size() added
// 12 Mar 2001 (TG) check_combination() added
//
// ********************************************************************
#include <iosfwd>
#include <list>
#include "pattern/CC_Flag.hh"
#include "gcombiner/GCombiner.hh"

class ToppIf;
template <class T>          class SmartPointer;
template <class C, class T> class GPattern;
template <class C, class T> class GPatternMatch;

/** A class which handles selection pattern for decaying particles.
    This class handles the kind of information which is independent from the
    Vertex object, like the Topp table entry of the decay and the list of track
    patterns which are used to select the tracks which will then be combined
    and vertexed. The list of pattern can be iterated with the
    GDecayPatternBase::iterator.

    @memo Generic class
    @author T. Glebe
*/
//==============================================================================
// GDecayPatternBase
//==============================================================================
template <class C, class T>
class GDecayPatternBase {
  typedef SmartPointer<GPattern<C,T> >  GPatPtr;
  typedef list<GPatPtr>                 GPat_l;
  typedef list<GPatPtr>::iterator       GPat_it;
  typedef list<GPatPtr>::const_iterator GPat_cit;
  //  typedef list<GPat_l*>                GPatlist_l;

#ifndef __CINT__
  typedef list<SmartPointer<GPatternMatch<C,T> > >     GPM_l;
  typedef GPM_l::iterator                   GPM_it;
  typedef GPM_l::const_iterator             GPM_cit;
  typedef GCombiner<GPM_l>                  PComb_v;
  typedef GCombiner<GPM_l>::iterator        PComb_it;
  typedef GC_trait<GPM_l>::GCombination_cit GCombination_cit;
#endif
  
public:
  /** @name --- Typedefs --- */
  ///
  typedef GPat_it      iterator;
  ///
  typedef GPat_cit     const_iterator;

  /** @name --- Constructors --- */
  ///
  GDecayPatternBase();
  ///
  GDecayPatternBase( const GDecayPatternBase<C,T>& rhs);
  ///
  GDecayPatternBase( const ToppIf& tp, const GDecayPatternBase<C,T>& rhs);
  ///
  GDecayPatternBase( const ToppIf& tp, 
		     const GPattern<C,T>& lhs, const GPattern<C,T>& rhs);
  ///
  virtual ~GDecayPatternBase();
  ///
  GDecayPatternBase<C,T>& operator=( const GDecayPatternBase<C,T>& rhs );

  /** @name --- Operators --- */
  ///
  GDecayPatternBase<C,T>& operator&( const GPattern<C,T>& rhs);
  ///  put everything into a single vertex (to be improved)
  GDecayPatternBase<C,T>& operator&( const GDecayPatternBase<C,T>& rhs);
  ///
  bool operator==( const ToppIf& rhs ) const;
  ///
//    GDecayPattern<C,T>& operator|( const GDecayPattern<C,T>& rhs) {
//      // rhs must be kept separate.
//      return *this;
//    }


  /** @name --- Access functions --- */
  ///
  inline const_iterator begin() const { return patlist.begin(); }
  ///
  inline const_iterator end()   const { return patlist.end(); }
  ///
  inline unsigned int   size()  const { return patlist.size(); }
  ///
  inline const ToppIf&  topp()  const { return *topp_; }
  /// set mass range cut around expected invariant mass
  void massrange(double mr);
  /// how to deal with charge conjugated tracks?
  void use_cc( CC_Flag cc ) const;
  /// get particle name
  string getName() const;
  ///
  string getDecayString() const;
  ///
  string getBasicDecayString(CC_Flag cc = NONE) const;
  ///
  string getPatternString() const;
  ///
  string getBasicPatternString(CC_Flag cc = NONE) const;
  
  /** @name --- Expert functions --- */
  ///
  virtual std::ostream& print(std::ostream& os) const;
  /// find out if there are pattern with equal charge
  bool isCollision() const;
  /// get a pattern by TOPP id
  const GPattern<C,T>* get(const ToppIf& topp) const;
  /// turn pattern into it's charge conjugated pendant
  const GDecayPatternBase<C,T>& flip_to_cc() const;
#ifndef __CINT__
  /// true if track combination passes pattern track cuts
  bool check_combination(const PComb_it& rhs) const;
#endif

protected:
  ///
  virtual void printPattern(std::ostream& os, CC_Flag cc = NONE) const;
  /// find out if pattern have same property
  bool eval_patlist();

  mutable const ToppIf* topp_;   // particle property
  mutable GPat_l        patlist; // list of pattern!
  mutable CC_Flag       useCC;   // indicate charge handling in vertex
  bool                  patsym;  // have pattern same property?
  double                dm;      // mass range of decaying particle
}; // end of GDecayPattern

#ifndef __CINT__
#include "pattern/GDecayPatternBase.icc"
#endif
#endif
