#ifndef __GDECAYPATTERNMATCH_HH
#define __GDECAYPATTERNMATCH_HH
// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   01. Feb 2001
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: class describing a decay pattern
//              match for Vt++ vertices
//
// changes:
// 05 Feb 2001 (TG) creation
// 07 Feb 2001 (TG) Changed Vertex & GDecayPattern pointer to SmartPointer.
// 26 Feb 2001 (TG) added argument to mass(), added operator==() & massdiff()
// 27 Feb 2001 (TG) added CheckCollision(), TrackCollision(), attatchVertex(),
//                  distToVertex(), valid(), valid_, PVdist, vtxobj
// 06 Mar 2001 (TG) CC_ added
// 08 Mar 2001 (TG) operator==(ToppIf), use_rm_, use_rm() added
// 09 Mar 2001 (TG) RefitVertex() added (mass constrained fit)
// 12 Mar 2001 (TG) getPattern() added
// 13 Mar 2001 (TG) renamed distToVertex() to TrackVertexDist()
// 13 Mar 2001 (TG) added VertexVertexDist(), ctau()
//
// ********************************************************************

#include <iosfwd>
#ifndef __CINT__
#include "pattern/SmartPointer.hh"
#include <list>
#endif

namespace VERTEX { 
  class Vertex;
  class Relation;
}
template <class C, class T> class GDecayPattern;
template <class C, class T> class GPatternMatch;
class VertexIf;
class ToppIf;
class PatternMatchIf;

/** Pattern Match class for Decays.
    GDecayPatternMatch represents a Vt++ vertex object together with the
    DecayPattern which was used to fit and select the vertex object.
    A list of track pattern matches is also kept to be able to create later
    relations between RecoVertex and RecoTrack objects.

    @memo Generic class
    @author T. Glebe
*/
//==============================================================================
// GDecayPatternMatch
//==============================================================================
template <class C, class T>
class GDecayPatternMatch {
public:
#ifndef __CINT__
  typedef list<SmartPointer<GPatternMatch<C,T> > >     GPM_l;
  typedef GPM_l::iterator       iterator;
  typedef GPM_l::const_iterator const_iterator;
#endif

  /** @name --- Constructors --- */
  ///
  GDecayPatternMatch();
  ///
  GDecayPatternMatch(const VERTEX::Vertex& v, const GDecayPattern<C,T>& vpat,
		     bool cc = false);
#ifndef __CINT__
  /// not available in CINT
  GDecayPatternMatch(const SmartPointer<VERTEX::Vertex>& v,
		     const SmartPointer<GDecayPattern<C,T> >& vpat, bool cc = false);
#endif
  ///
  virtual ~GDecayPatternMatch();

  /** @name --- Access fuctions --- */
  ///
  inline const VERTEX::Vertex& getVertex() const { return *vtx; }
  ///
  inline VERTEX::Vertex& getVertex() { return *vtx; }
  ///
  inline const GDecayPattern<C,T>& getPattern() const { return *vtxpat; }
  ///
  inline const VertexIf& getAttachedVertex() const { return *vtxobj; }
  /// vtx mass. usepat=false: calc vtx mass from GPattern rest-masses
  double mass() const;
  /// $|m - m_{topp}|$
  double massdiff() const;
  /// return validity
  inline bool valid() const { return valid_; }
  /// mark match as invalid
  inline void invalid() { vtx->set_invalid(); valid_ = false; }
  /// attatch a vertex from src container
  template <class InputContainer>
  bool attachVertex(const InputContainer& src);
  /// min. distance of reconstructed track to vertex object (e.g. primary Vertex)
  inline double TrackVertexDist() const { return PVdist; }
  /// spatial distance of vertex to associated vertex
  double VertexVertexDist() const;
  /// $c\tau = | |\vec{x}_1 - \vec{x}_2| \cdot m / p_{tot}|$
  double ctau() const { return fabs(VertexVertexDist() * mass() / vtx->p()); }
  /** Max impact of an outgoing track to attached vertex. Good to check whether
      vertex is displaced. */
  double maxImpact() const;
  /// is the charge conjugated pattern matched?
  inline bool isCC() const { return CC_; }
  /// use rest-masses from pattern matches?
  inline bool use_rm() const { return use_rm_; }
  /// set rest-mass usage
  inline void use_rm(bool use) { use_rm_ = use; }

  /** @name --- Operators --- */
  ///
  bool operator==(const GDecayPatternMatch<C,T>& rhs) const;
  ///
  bool operator==(const ToppIf& rhs) const;

  /** @name --- Expert functions --- */
  ///
  std::ostream& print(std::ostream& os);
  /// check whether rhs has one or more track in common
  bool TrackCollision(const GDecayPatternMatch<C,T>& rhs) const;
  ///
  bool CheckCollision() const;

#ifndef __CINT__
  /// get a pattern match by TOPP id
  const GPatternMatch<C,T>* get(const ToppIf& topp) const;
  ///
  const VERTEX::Relation* getRelation(const ToppIf& topp) const;

  /// not avialable in CINT
  void push_back(const SmartPointer<GPatternMatch<C,T> >& ptr);
  /// not avialable in CINT
  void push_back(const GPatternMatch<C,T>& rhs);
  ///
  inline       iterator begin()       { return pmlist.begin(); }
  ///
  inline       iterator end()         { return pmlist.end(); }
  ///
  inline const_iterator begin() const { return pmlist.begin(); }
  ///
  inline const_iterator end()   const { return pmlist.end(); }
  ///
  inline unsigned int   size()  const { return pmlist.size(); }
#endif

  /// update pattern matches
  void update_pattern_matches();
  /// do mass constrained vertex fit
  bool RefitVertex();

protected:
#ifndef __CINT__
  SmartPointer<VERTEX::Vertex>       vtx;     // not persistent up to now!
  SmartPointer<GDecayPattern<C,T> >  vtxpat;  // decay pattern
  GPM_l                              pmlist;  // list of pattern matches
#endif
  mutable bool                       valid_;  // mark match as invalid
  bool                               CC_   ;  // match on charge conjugated pattern?
  bool                               use_rm_; // use rest-masses from pattern matches?
  double                             PVdist;  // min. dist to Primary Vertex
  VertexIf*                          vtxobj;  // vertex to associate with
}; // end of GDecayPatternMatch

#ifndef __CINT__
#include "pattern/GDecayPatternMatch.icc"
#endif
#endif
