#ifndef __GPATTERN_HH
#define __GPATTERN_HH
// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   23. Jan 2001
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: A new Pattern class (to merge ClPdtEntry and Pattern)
//
// changes:
// 23 Jan 2001 (TG) creation
// 01 Feb 2001 (TG) added cprop, useCC, set_cprop(), use_cc() from
//                  Pattern class
// 07 Feb 2001 (TG) renamed topp to topp_, added topp() member function
// 09 Feb 2001 (TG) added exist() member
// 02 Mar 2001 (TG) operator==() added
// 05 Mar 2001 (TG) getString(), getName(), flip_to_cc(), PropertyString() added
// 06 Mar 2001 (TG) operator==(GPattern) added, renamed property to property_,
//                  added property(), check_property(), toppCC()
// 08 Mar 2001 (TG) CTOR added
//
// ********************************************************************

#include <iosfwd>
#include "pattern/CC_Flag.hh"
#include <string>

class ToppIf;
template <class C, class T> class ClProperty;
template <class C, class T> class GDecayPattern;

/** A generic template class for defining object selection pattern, based on
    ClProperty expressions.

    @memo Generic class
    @author T. Glebe
 */
//==============================================================================
// GPattern
//==============================================================================
template <class C, class T>
class GPattern {
public:
  /** @name --- Constructors --- */
  ///
  GPattern();
  ///
  GPattern(const GPattern<C,T>& rhs);
  ///
  GPattern(const ToppIf& gt);
  ///
  GPattern(const ToppIf& gt, const ClProperty<C,T>& rhs);
  ///
  virtual ~GPattern();
  ///
  GPattern<C,T>& operator=(const GPattern<C,T>& rhs);

  /** @name --- Operators --- */
  /// OR of two patterns
  GPattern<C,T> operator|( const GPattern<C,T>& rhs ) const;
  /// AND of patterns defines a decay
  GDecayPattern<C,T> operator&( const GPattern<C,T>& rhs) const;
  ///
  bool operator==(const ToppIf& rhs) const;
  ///
  bool operator==(const GPattern<C,T>& rhs) const;

  /** @name --- User functions --- */
  /// check if single Track passes property criterias
  virtual T evalTrack(const C& rhs) const { return cprop->evaluate(rhs); }
  /** select all tracks from InputContainer which passes property 
      criteria and put them as PatternMatches in OutputContainer */
  template <class InputContainer, class OutputContainer>
  bool select(const InputContainer&  src, OutputContainer& trg) const;
  /// return true if a track which matches the pattern exists in InputContainer
  template <class InputContainer>
  bool exist(const InputContainer& src) const;

  /// set cprop pointer
  void set_cprop(const ClProperty<C,T>& rhs);
  /// steers treatment of charge
  void use_cc(CC_Flag flag);
  /// get topp object
  inline const ToppIf& topp() const { return *topp_; }
  /// get charge conjugated topp object
  inline const ToppIf& toppCC() const { return *toppCC_; }
  /// get property object
  inline const ClProperty<C,T>& property() const {
    return *property_;
  }
  /// get pattern Name
  string getName() const;
  /// get property string
  inline string PropertyString() const { return property().getString(); }
  /// get the pattern string
  inline string getString() const {
    return (property_!=0) ? getName() + string(" with ") + property_->getString() :
      getName();
  }
  /// true if rhs is equal to pattern property
  bool check_property(const ClProperty<C,T>& rhs) const;

  /** @name --- Expert functions --- */
  /// set charge conjugated particle hypothesis
  bool set_cc(const ToppIf& rhs);
  /// transform pattern into is't charge conjugated
  GPattern<C,T>& flip_to_cc();
  /// used by operator<<
  virtual std::ostream& print(std::ostream& os) const;
  /// merge two pattern, only merge property
  void merge_property(const GPattern<C,T>& rhs);

protected:
  ///
  void set_property(const ClProperty<C,T>& rhs);

  const ToppIf*          topp_;     // particle property
  const ToppIf*          toppCC_;   // charge conjugated hypothesis
  ClProperty<C,T>*       property_; // cuts defined by ClProperty
  ClProperty<C,T>*       cprop;     // cuts plus charge cut
  CC_Flag                useCC;     // indicate charge handling
}; // end of GPattern.hh


//==============================================================================
// operator <<: the second template argument is a problem, therefore hardcoded
//==============================================================================
template <class T>
inline std::ostream& operator<<( std::ostream& os, const GPattern<T,int>& a ) {
  return a.print( os );
}

//  template <class T>
//  inline std::ostream& operator<<( std::ostream& os, const GPattern<T,float>& a ) {
//    return a.print( os );
//  }

#ifndef __CINT__
#include "pattern/GPattern.icc"
#endif
#endif
