#ifndef __GPATTERNMATCH_HH
#define __GPATTERNMATCH_HH
// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   23. Jan 2001
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: A new PatternMatch class (replacement of PatternMatch)
//
// changes:
// 23 Jan 2001 (TG) creation
// 26 Feb 2001 (TG) operator==() added
// 28 Feb 2001 (TG) operator==(ToppIf) added
// 28 Feb 2001 (TG) added inheritance from PatternMatchIf
// 05 Mar 2001 (TG) cloneCC() added
// 09 Mar 2001 (TG) setPattern() added
//
// ********************************************************************

#include <iosfwd>
#ifndef __CINT__
#include "pattern/SmartPointer.hh"
#endif
#include "interfaces/PatternMatchIf.hh"

template <class C, class T> class GPattern;
namespace VERTEX { class Track; }
class TrackIf;
class ToppIf;

/** Class which represents a selected track object together with the cuts which
    were used to select it. The cuts are represented by a ClProperty
    expression object.

    @memo Generic class
    @author T. Glebe
*/
//==============================================================================
// GPatternMatch
//==============================================================================
template <class C, class T>
class GPatternMatch : public PatternMatchIf {
public:
  /** @name --- Constructors --- */
  ///
  GPatternMatch();
  ///
  GPatternMatch(const C& tr, const GPattern<C,T>& pat);
#ifndef __CINT__
  /// not available in CINT
  GPatternMatch(const C& tr, const SmartPointer<GPattern<C,T> >& pat);
#endif
  ///
  virtual ~GPatternMatch();

  /** @name --- Access methods --- */
  /// read only track access on interface-level
  const TrackIf& getTrackIf() const { return *track; }
  /// read only track access to full track
  inline const C& getTrack() const { return *track; }
  /// read only pattern access
  inline const GPattern<C,T>& getPattern() const { return *pattern; }
  /// return Vt track
  inline const VERTEX::Track& getVtTrack() const { return *vttrack; }
  /// return Vt track
  inline VERTEX::Track& getVtTrack() { return *vttrack; }

  /** @name --- Operators --- */
  ///
  bool operator==(const GPatternMatch<C,T>& rhs) const;
  ///
  bool operator==(const ToppIf& rhs) const;

  /** @name --- Expert functions --- */
  ///
  std::ostream& print(std::ostream& os) const;
  /// return pattern match clone
  GPatternMatch<C,T> clone() const;
  /// return charge conjugated pattern match clone
  GPatternMatch<C,T> cloneCC() const;
#ifndef __CINT__
  /// set a new Pattern
  inline void setPattern(const SmartPointer<GPattern<C,T> >& rhs) { pattern = rhs; }
#endif

private:
  const C*                           track;    // track object
  //  const TrackIf*                     track;    // track object
#ifndef __CINT__
  SmartPointer<GPattern<C,T> >       pattern;   // pattern which selected track
  SmartPointer<VERTEX::Track>        vttrack;   // track of Vt vertexing code
#endif
}; // end of GPatternMatch

//==============================================================================
// operator <<: the second template argument is a problem, therefore hardcoded
//==============================================================================
template <class T>
inline std::ostream& operator<<( std::ostream& os, const GPatternMatch<T,int>& a ) {
  return a.print( os );
}

#ifndef __CINT__
#include "pattern/GPatternMatch.icc"
#endif
#endif
