#ifndef __SDecayPatternCC_HH
#define __SDecayPatternCC_HH
// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   20. Apr 2001
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Class describing a decay pattern plus its charge conjugated.
//
// changes:
// 23 Apr 2001 (TG) creation
//
// ********************************************************************
#include <iosfwd>
#include "pattern/CC_Flag.hh"

class ToppIf;
template <class T>          class SmartPointer;
template <class C, class T> class SPattern;
template <class C, class T> class SPatternMatch;

/** A class which handles selection pattern for decaying particles.
    This class handles the kind of information which is independent from the
    Vertex object, like the Topp table entry of the decay and the list of track
    patterns which are used to select the tracks which will then be combined
    and vertexed.

    @memo Generic class
    @author T. Glebe
*/
//==============================================================================
// SDecayPatternCC
//==============================================================================
template <class C, class T, unsigned int NTR>
class SDecayPatternCC {
  typedef SmartPointer<SPattern<C,T> >  SPatPtr;

public:
  /** @name --- Constructors --- */
  ///
  SDecayPatternCC();
  ///
  SDecayPatternCC( const SDecayPatternCC<C,T,NTR>& rhs);
  ///
  SDecayPatternCC( const SDecayPatternCC<C,T,NTR-1>& lhs, const SPatPtr& rhs);
  ///
  SDecayPatternCC( const ToppIf& tp, const SDecayPatternCC<C,T,NTR>& rhs);
  ///
  SDecayPatternCC( const ToppIf& tp, 
		 const SPatPtr& lhs,
		 const SPatPtr& rhs);
  ///
  virtual ~SDecayPatternCC();
  ///
  SDecayPatternCC<C,T,NTR>& operator=( const SDecayPatternCC<C,T,NTR>& rhs );

  /** @name --- Operators --- */
  ///
  SDecayPatternCC<C,T,NTR+1> operator&( const SPatPtr& rhs) const;
  ///
  bool operator==( const ToppIf& rhs ) const;

  /** @name --- Access functions --- */
  ///
  static const unsigned int size = NTR;
  ///
  inline const ToppIf&  topp()  const { return *topp_; }
  /// read-only pattern access
  inline const SPatPtr& pattern(unsigned int i) const { return patlist[i]; }
  /// read/write pattern access
  inline SPatPtr& pattern(unsigned int i) { return patlist[i]; }
  /// set mass range cut around expected invariant mass
  void massrange(double mr);
  /// how to deal with charge conjugated tracks?
  void use_cc( CC_Flag cc ) const;
  /// get particle name
  string getName() const;
  /// get the pattern string
  string getString() const;
  ///
  string getDecayString() const;
  ///
  string getBasicDecayString(CC_Flag cc = NONE) const;
  ///
  string getPatternString() const;
  ///
  string getBasicPatternString(CC_Flag cc = NONE) const;
  /// vector with rest-masses
  inline const SVector<double,NTR>& rm()   const { return rm_; }
  /// vector with charge conjugated rest-masses
  inline const SVector<double,NTR>& rmCC() const { return rmCC_; }
  
  /** @name --- Expert functions --- */
  ///
  virtual std::ostream& print(std::ostream& os) const;
  /// find out if there are pattern with equal charge
  bool isCollision() const;
  /// get a pattern by TOPP id
  const SPatPtr get(const ToppIf& topp) const;
  /// turn pattern into it's charge conjugated pendant
  const SDecayPatternCC<C,T,NTR>& flip_to_cc() const;
#ifdef XXX
  /// true if track combination passes pattern track cuts
  bool check_combination(const PComb_it& rhs) const;
#endif

protected:
  ///
  virtual void printPattern(std::ostream& os, CC_Flag cc = NONE) const;
  /// find out if pattern have same property
  bool eval_patlist();

  mutable const ToppIf* topp_;   // particle property
  mutable SPatPtr patlist[NTR];  // list of pattern pointer!
  mutable CC_Flag       useCC;   // indicate charge handling in vertex
  bool                  patsym;  // have pattern same property?
  double                dm;      // mass range of decaying particle
  SVector<double,NTR> rm_;       // vector of restmasses
  SVector<double,NTR> rmCC_;     // vector of charge conjugated restmasses
}; // end of SDecayPatternCC

//==============================================================================
// operator<<
//==============================================================================
template <class T, unsigned int NTR>
inline std::ostream& operator<<(std::ostream& os,
				const SDecayPatternCC<RecoTrack,T,NTR>& rhs) {
  return rhs.print(os);
}

template <class T, unsigned int NTR>
inline std::ostream& operator<<(std::ostream& os,
				const SmartPointer<SDecayPatternCC<RecoTrack,T,NTR> >& rhs) {
  return rhs->print(os);
}

#include "pattern/SDecayPatternCC.icc"
#endif
